# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler for built-in collection types (lists, sets, dicts, etc)."""

from __future__ import annotations

import dataclasses
from typing import Any, Callable, Optional, Sequence

from treescope import dataclass_util
from treescope import layout_algorithms
from treescope import renderers
from treescope._internal.parts import basic_parts
from treescope._internal.parts import common_structures
from treescope._internal.parts import common_styles
from treescope._internal.parts import part_interface


CSSStyleRule = part_interface.CSSStyleRule
HtmlContextForSetup = part_interface.HtmlContextForSetup


def build_field_children(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
    fields_or_attribute_names: Sequence[dataclasses.Field[Any] | str],
    attr_style_fn: (
        Callable[[str], part_interface.RenderableTreePart] | None
    ) = None,
) -> list[part_interface.RenderableTreePart]:
  """Renders a set of fields/attributes into a list of comma-separated children.

  This is a helper function used for rendering dataclasses, namedtuples, and
  similar objects, of the form ::

    ClassName(
        field_name_one=value1,
        field_name_two=value2,
    )

  If `fields_or_attribute_names` includes dataclass fields:

  * Metadata for the fields will be visible on hover,

  * Fields with ``repr=False`` will be hidden unless roundtrip mode is enabled.

  Args:
    node: Node to render.
    path: Path to this node.
    subtree_renderer: How to render subtrees (see `TreescopeSubtreeRenderer`)
    fields_or_attribute_names: Sequence of fields or attribute names to render.
      Any field with the metadata key "treescope_always_collapse" set to True
      will always render collapsed.
    attr_style_fn: Optional function which makes attributes to a part that
      should render them. If not provided, all parts are rendered as plain text.

  Returns:
    A list of child objects. This can be passed to
    `common_structures.build_foldable_tree_node_from_children` (with
    ``comma_separated=False``)
  """
  if attr_style_fn is None:
    attr_style_fn = basic_parts.Text

  field_names = []
  fields: list[Optional[dataclasses.Field[Any]]] = []
  for field_or_name in fields_or_attribute_names:
    if isinstance(field_or_name, str):
      field_names.append(field_or_name)
      fields.append(None)
    else:
      field_names.append(field_or_name.name)
      fields.append(field_or_name)

  children = []
  for i, (field_name, maybe_field) in enumerate(zip(field_names, fields)):
    child_path = None if path is None else f"{path}.{field_name}"

    if i < len(fields) - 1:
      # Not the last child. Always show a comma, and add a space when
      # collapsed.
      comma_after = basic_parts.siblings(
          ",", basic_parts.FoldCondition(collapsed=basic_parts.Text(" "))
      )
    else:
      # Last child: only show the comma when the node is expanded.
      comma_after = basic_parts.FoldCondition(expanded=basic_parts.Text(","))

    if maybe_field is not None:
      hide_except_in_roundtrip = not maybe_field.repr
      force_collapsed = maybe_field.metadata.get(
          "treescope_always_collapse", False
      )
    else:
      hide_except_in_roundtrip = False
      force_collapsed = False

    field_name_rendering = attr_style_fn(field_name)

    try:
      field_value = getattr(node, field_name)
    except AttributeError:
      child = basic_parts.FoldCondition(
          expanded=common_styles.CommentColor(
              basic_parts.siblings("# ", field_name_rendering, " is missing")
          )
      )
    else:
      child = basic_parts.siblings_with_annotations(
          field_name_rendering,
          "=",
          subtree_renderer(field_value, path=child_path),
      )

    child_line = basic_parts.build_full_line_with_annotations(
        child, comma_after
    )
    if force_collapsed:
      layout_algorithms.expand_to_depth(child_line, 0)
    if hide_except_in_roundtrip:
      child_line = basic_parts.RoundtripCondition(roundtrip=child_line)

    children.append(child_line)

  return children


def render_dataclass_constructor(
    node: Any,
) -> part_interface.RenderableTreePart:
  """Renders the constructor for a dataclass, including the open parenthesis."""
  assert dataclasses.is_dataclass(node) and not isinstance(node, type)
  if not dataclass_util.init_takes_fields(type(node)):
    constructor_open = basic_parts.siblings(
        basic_parts.RoundtripCondition(
            roundtrip=basic_parts.Text(
                "treescope.dataclass_util.dataclass_from_attributes("
            )
        ),
        common_structures.maybe_qualified_type_name(type(node)),
        basic_parts.RoundtripCondition(
            roundtrip=basic_parts.Text(", "),
            not_roundtrip=basic_parts.Text("("),
        ),
    )
  else:
    constructor_open = basic_parts.siblings(
        common_structures.maybe_qualified_type_name(type(node)), "("
    )
  return constructor_open



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configures the default renderer, and allows reconfiguring it dynamically."""

import contextlib
import functools
from typing import Any, Callable
from treescope import context
from treescope import handlers
from treescope import layout_algorithms
from treescope import lowering
from treescope import renderers
from treescope import rendering_parts
from treescope import type_registries


active_renderer: context.ContextualValue[renderers.TreescopeRenderer] = (
    context.ContextualValue(
        module=__name__,
        qualname="active_renderer",
        initial_value=renderers.TreescopeRenderer(
            handlers=[
                # Objects with `__treescope_repr__` defined.
                handlers.handle_via_treescope_repr_method,
                # Objects in the global registry of custom type handlers.
                handlers.handle_via_global_registry,
                # Reflection of functions and classes.
                handlers.handle_code_objects_with_reflection,
                # Render basic builtin types, namedtuples, and dataclasses.
                handlers.handle_basic_types,
                # Fallback for unknown pytree types: Show repr and also the
                # PyTree children. Note: This is a no-op unless JAX has been
                # imported.
                handlers.handle_arbitrary_pytrees,
                # Fallback to ordinary `repr` for any other object.
                handlers.handle_anything_with_repr,
            ],
            wrapper_hooks=[
                # Allow user-configurable visualizations.
                handlers.use_autovisualizer_if_present,
                # Show display objects inline.
                handlers.append_repr_html_when_present,
                # Collapse well-known objects into aliases.
                handlers.replace_with_canonical_aliases,
                # Annotate multiple references to the same mutable Python
                # object.
                handlers.check_for_shared_values,
            ],
            context_builders=[
                # Set up a new context for each rendered object when rendering
                # shared values.
                handlers.setup_shared_value_context,
                # Update type registries to account for newly imported
                # modules before rendering.
                type_registries.update_registries_for_imports,
            ],
        ),
    )
)


active_expansion_strategy = context.ContextualValue[
    Callable[[rendering_parts.RenderableTreePart], None]
](
    module=__name__,
    qualname="active_expansion_strategy",
    initial_value=layout_algorithms.expand_for_balanced_layout,
)


def using_expansion_strategy(
    max_height: int | None = 20,
    target_width: int = 60,
    relax_for_only_child: bool = True,
    recursive_expand_height: int = 7,
) -> contextlib.AbstractContextManager[None]:
  """Context manager that modifies the expansion strategy of Treescope.

  This is a convenience function for modifying the layout algorithm used by
  Treescope to determine how many nodes to expand by default. It returns a
  context manager within which the given balanced expansion strategy will be
  used.

  Args:
    max_height: Maximum number of lines for the rendering, which we should try
      not to exceed when expanding nodes. Can sometimes be exceeded for "only
      child" subtrees if ``relax_for_only_child`` is set. If None, then an
      arbitrary height is allowed, and only ``target_width`` will be used to
      determine expansion size.
    target_width: Target number of characters for the rendering. If a node is
      shorter than this width when collapsed, we will leave it on a single line
      instead of expanding it.
    relax_for_only_child: If True, we allow the height constraint to be violated
      once if there is only a single node that could possibly be expanded. In
      particular, we repeatedly expand nodes until we reach a node shorter than
      ``target_width`` or a node with two or more children. If we match or
      exceed the height constraint while doing this, we stop after expanding
      that node (but leave it expanded). If we haven't reached the constraint,
      we continue expanding under the stricter rules.
    recursive_expand_height: Whenever we mark a node as collapsed, we first
      recursively expand its children as if it was expanded with this as the max
      height, then collapse only the outermost node. This means it will render
      as a single line, but if expanded by the user, it will expand not just the
      clicked node but also some of its children. This is intended to support
      faster interactive exploration of large objects.

  Returns:
    Context manager in which the given strategy will be active.
  """
  return active_expansion_strategy.set_scoped(
      functools.partial(
          layout_algorithms.expand_for_balanced_layout,
          max_height=max_height,
          target_width=target_width,
          relax_height_constraint_for_only_child=relax_for_only_child,
          recursive_expand_height_for_collapsed_nodes=recursive_expand_height,
      )
  )


def build_foldable_representation(
    value: Any,
    ignore_exceptions: bool = False,
) -> rendering_parts.RenderableAndLineAnnotations:
  """Builds a foldable representation of an object using default configuration.

  Uses the default renderer and expansion strategy.

  Args:
    value: Value to render.
    ignore_exceptions: Whether to catch errors during rendering of subtrees and
      show a fallback for those subtrees, instead of failing the entire
      renderer. Best used in contexts where `value_to_foldable_html` is not
      being called directly, e.g. when registering this as a default
      pretty-printer.

  Returns:
    A text representation of the object.
  """
  foldable_ir_with_annotations = (
      active_renderer.get().to_foldable_representation(
          value, ignore_exceptions=ignore_exceptions
      )
  )
  # Expand the renderable ignoring its annotations.
  active_expansion_strategy.get()(foldable_ir_with_annotations.renderable)
  return foldable_ir_with_annotations


def render_to_text(
    value: Any,
    roundtrip_mode: bool = False,
    ignore_exceptions: bool = False,
) -> str:
  """Renders an object to text using the default renderer.

  Args:
    value: Value to render.
    roundtrip_mode: Whether to render in roundtrip mode.
    ignore_exceptions: Whether to catch errors during rendering of subtrees and
      show a fallback for those subtrees, instead of failing the entire
      renderer.

  Returns:
    A text representation of the object.
  """
  foldable_ir = rendering_parts.build_full_line_with_annotations(
      build_foldable_representation(value, ignore_exceptions=ignore_exceptions)
  )
  return lowering.render_to_text_as_root(foldable_ir, roundtrip_mode)


def render_to_html(
    value: Any,
    roundtrip_mode: bool = False,
    ignore_exceptions: bool = False,
    compressed: bool = True,
) -> str:
  """Renders an object to HTML using the default renderer.

  Args:
    value: Value to render.
    roundtrip_mode: Whether to render in roundtrip mode.
    ignore_exceptions: Whether to catch errors during rendering of subtrees and
      show a fallback for those subtrees, instead of failing the entire
      renderer.
    compressed: Whether to compress the output HTML.

  Returns:
    HTML source code for the foldable representation of the object.
  """
  foldable_ir = rendering_parts.build_full_line_with_annotations(
      build_foldable_representation(value, ignore_exceptions=ignore_exceptions)
  )
  return lowering.render_to_html_as_root(
      foldable_ir, roundtrip_mode, compressed=compressed
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parts and builders for Treescope's intermediate output format.

The functions exposed here can be used to construct a tree of parts that can be
rendered to text or to interactive HTML. Node handlers and the
__treescope_repr__ method can use them to render custom types.

Note that the internal definition of `RenderableTreePart` is considered an
implementation detail, and is subject to change. To build renderable tree parts,
you should instead use the functions exposed here (or higher-level wrappers in
`treescope.repr_lib`).
"""

# pylint: disable=g-importing-member,g-multiple-import,unused-import


from treescope._internal.parts.basic_parts import (
    build_full_line_with_annotations,
    empty_part,
    floating_annotation_with_separate_focus,
    fold_condition,
    in_outlined_box,
    indented_children,
    on_separate_lines,
    roundtrip_condition,
    siblings_with_annotations,
    siblings,
    summarizable_condition,
    text,
    vertical_space,
    with_layout_mark,
)
from treescope._internal.parts.common_structures import (
    build_copy_button,
    build_custom_foldable_tree_node,
    build_foldable_tree_node_from_children,
    build_one_line_tree_node,
    maybe_qualified_type_name,
)
from treescope._internal.parts.common_styles import (
    abbreviation_color,
    comment_color_when_expanded,
    comment_color,
    custom_text_color,
    deferred_placeholder_style,
    error_color,
    qualified_type_name_style,
    custom_style,
)
from treescope._internal.parts.custom_dataclass_util import (
    build_field_children,
    render_dataclass_constructor,
)
from treescope._internal.parts.embedded_iframe import (
    embedded_iframe,
)
from treescope._internal.parts.part_interface import (
    RenderableTreePart,
    RenderableAndLineAnnotations,
    Rendering,
    ExpandState,
    NodePath,
)


__all__ = [
    "build_full_line_with_annotations",
    "empty_part",
    "floating_annotation_with_separate_focus",
    "fold_condition",
    "in_outlined_box",
    "indented_children",
    "on_separate_lines",
    "roundtrip_condition",
    "siblings_with_annotations",
    "siblings",
    "summarizable_condition",
    "text",
    "vertical_space",
    "with_layout_mark",
    "build_copy_button",
    "build_custom_foldable_tree_node",
    "build_foldable_tree_node_from_children",
    "build_one_line_tree_node",
    "maybe_qualified_type_name",
    "abbreviation_color",
    "comment_color_when_expanded",
    "comment_color",
    "custom_text_color",
    "deferred_placeholder_style",
    "error_color",
    "qualified_type_name_style",
    "custom_style",
    "build_field_children",
    "render_dataclass_constructor",
    "embedded_iframe",
    "RenderableTreePart",
    "RenderableAndLineAnnotations",
    "Rendering",
    "ExpandState",
    "NodePath",
]



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for NDArray adapters and array visualization."""

from absl.testing import absltest
from absl.testing import parameterized
import jax.numpy as jnp
import numpy as np
import torch
import treescope
from treescope import ndarray_adapters
from treescope import type_registries


class NdarrayAdaptersTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    type_registries.update_registries_for_imports()

  @parameterized.product(
      array_type=["jax", "torch"],
      dtype=[np.int32, np.float32, np.bool_],
  )
  def test_adapter_positional_numpy_consistency(self, array_type, dtype):

    reshaped_arange = np.arange(19 * 23).reshape((19, 23))

    if dtype == np.bool_:
      array_np = (reshaped_arange % 2) == 0
    else:
      array_np = reshaped_arange.astype(dtype)

    mask_np = (reshaped_arange % 3) != 0

    if array_type == "jax":
      array = jnp.array(array_np)
      mask = jnp.array(mask_np)
    elif array_type == "torch":
      array = torch.tensor(array_np)
      mask = torch.tensor(mask_np)
    else:
      raise ValueError(f"Unsupported array_type: {array_type}")

    np_adapter = type_registries.lookup_ndarray_adapter(array_np)
    self.assertIsNotNone(np_adapter)

    cur_adapter = type_registries.lookup_ndarray_adapter(array)
    self.assertIsNotNone(cur_adapter)

    with self.subTest("axis_info"):
      self.assertEqual(
          np_adapter.get_axis_info_for_array_data(array_np),
          cur_adapter.get_axis_info_for_array_data(array),
      )

    with self.subTest("data_untruncated_unmasked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=None, edge_items_per_axis=(None, None)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=None, edge_items_per_axis=(None, None)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_untruncated_masked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=mask_np, edge_items_per_axis=(None, None)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=mask, edge_items_per_axis=(None, None)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_semitruncated_unmasked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=None, edge_items_per_axis=(2, None)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=None, edge_items_per_axis=(2, None)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_semitruncated_unmasked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=mask_np, edge_items_per_axis=(2, None)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=mask, edge_items_per_axis=(2, None)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_truncated_unmasked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=None, edge_items_per_axis=(2, 4)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=None, edge_items_per_axis=(2, 4)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_truncated_unmasked"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=mask_np, edge_items_per_axis=(2, 4)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=mask, edge_items_per_axis=(2, 4)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    with self.subTest("data_truncated_broadcast_mask"):
      data_out_np, mask_out_np = np_adapter.get_array_data_with_truncation(
          array=array_np, mask=mask_np[0, :], edge_items_per_axis=(3, 7)
      )
      data_out, mask_out = cur_adapter.get_array_data_with_truncation(
          array=array, mask=mask[0, :], edge_items_per_axis=(3, 7)
      )
      np.testing.assert_array_equal(data_out_np, data_out)
      np.testing.assert_array_equal(mask_out_np, mask_out)

    for fast in (True, False):
      with self.subTest("summary_fast" if fast else "summary_slow"):
        summary_info_np = np_adapter.get_array_summary(array_np, fast=True)
        summary_info = cur_adapter.get_array_summary(array, fast=True)
        self.assertEqual(
            summary_info_np.split(" ", 1)[1], summary_info.split(" ", 1)[1]
        )

  def test_pytorch_named_axes_info(self):
    data = np.arange(19 * 23).reshape((19, 23))
    array = torch.tensor(data).rename("foo", None)
    adapter = type_registries.lookup_ndarray_adapter(array)
    self.assertIsNotNone(adapter)
    self.assertEqual(
        adapter.get_axis_info_for_array_data(array),
        (
            ndarray_adapters.NamedPositionalAxisInfo(0, "foo", 19),
            ndarray_adapters.PositionalAxisInfo(1, 23),
        ),
    )

  @parameterized.product(
      array_type=["numpy", "jax", "torch"],
      dtype=[np.int32, np.float32, np.bool_],
  )
  def test_array_rendering_without_error(self, array_type, dtype):
    reshaped_arange = np.arange(19 * 23).reshape((19, 23))

    if dtype == np.bool_:
      array_np = (reshaped_arange % 2) == 0
    else:
      array_np = reshaped_arange.astype(dtype)

    if array_type == "numpy":
      array = array_np
    elif array_type == "jax":
      array = jnp.array(array_np)
    elif array_type == "torch":
      array = torch.tensor(array_np)
    else:
      raise ValueError(f"Unsupported array_type: {array_type}")

    with self.subTest("explicit_unmasked"):
      res = treescope.render_array(array)
      self.assertTrue(hasattr(res, "_repr_html_"))

    with self.subTest("explicit_masked"):
      res = treescope.render_array(array, valid_mask=array > 100)
      self.assertTrue(hasattr(res, "_repr_html_"))

    with self.subTest("explicit_masked_truncated"):
      res = treescope.render_array(
          array, valid_mask=array > 100, truncate=True, maximum_size=100
      )
      self.assertTrue(hasattr(res, "_repr_html_"))

    with self.subTest("automatic"):
      with treescope.active_autovisualizer.set_scoped(
          treescope.ArrayAutovisualizer()
      ):
        res = treescope.render_to_html(
            array, ignore_exceptions=False, compressed=False
        )
        self.assertIsInstance(res, str)
        self.assertIn("arrayviz", res)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handlers for certain "well-known" objects, such as module functions.

Taking the `repr` of a function or callable usually produces something like

  <function vmap at 0x7f98bf556440>

and in some cases produces something like

  <jax.custom_derivatives.custom_jvp object at 0x7f98c0b5f130>

This can make it hard to determine what object this actually is from a user
perspective, and it would likely be more user-friendly to just output `jax.vmap`
or `jax.nn.relu` as the representation of this object.

See `canonical_aliases` for details of how we determine aliases for a subset
of common functions.
"""

from __future__ import annotations

from typing import Any

from treescope import canonical_aliases
from treescope import renderers
from treescope import rendering_parts


def replace_with_canonical_aliases(
    node: Any,
    path: str | None,
    node_renderer: renderers.TreescopeSubtreeRenderer,
    summarization_threshold: int = 20,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Rewrites objects to use well-known aliases when known.

  This rendering hook checks if the object is a well-known object exported by
  a public API, and if so, makes its short representation be its path in that
  API rather than its repr.

  Objects are looked up in canonical_aliases to determine if there is a
  specific alias registered for them. If not, but the object defines
  `__module__` and `__qualname__` attributes, these are used to determine the
  canonical alias instead.

  For objects whose fully-qualified name is longer than
  `summarization_threshold`, we shorten it when collapsed or not in roundtrip
  mode.

  Args:
    node: The node that has been rendered
    path: Optionally, a path to this node as a string.
    node_renderer: The inner renderer for this node. This should be used to
      render `node` itself into HTML tags.
    summarization_threshold: Threshold at which to omit the fully-qualified name
      except in roundtrip mode.

  Returns:
    A possibly-modified representation of this object, or NotImplemented.
  """
  maybe_alias = canonical_aliases.lookup_alias(
      node, infer_from_attributes=True, allow_relative=True
  )
  if not maybe_alias:
    # No alias.
    return NotImplemented

  qualified_name = str(maybe_alias)
  if "." not in qualified_name:
    name_rendering = qualified_name
  else:
    prefix, suffix = qualified_name.rsplit(".", 1)
    if len(qualified_name) > summarization_threshold:
      name_rendering = rendering_parts.siblings(
          rendering_parts.summarizable_condition(
              detail=rendering_parts.qualified_type_name_style(
                  rendering_parts.text(prefix + ".")
              )
          ),
          suffix,
      )
    else:
      name_rendering = rendering_parts.siblings(
          rendering_parts.qualified_type_name_style(
              rendering_parts.text(prefix + ".")
          ),
          suffix,
      )

  original_rendering = node_renderer(node, path=path)

  return rendering_parts.build_custom_foldable_tree_node(
      label=rendering_parts.comment_color_when_expanded(
          rendering_parts.siblings(
              rendering_parts.fold_condition(
                  expanded=rendering_parts.text("# ")
              ),
              name_rendering,
          )
      ),
      contents=rendering_parts.fold_condition(
          expanded=rendering_parts.indented_children([original_rendering])
      ),
      # Aliases should start collapsed regardless of layout.
      expand_state=rendering_parts.ExpandState.COLLAPSED,
      path=path,
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for using treescope in IPython."""

import contextlib
from typing import Any

from treescope import context
from treescope import figures
from treescope import lowering
from treescope import rendering_parts
from treescope._internal import object_inspection
from treescope._internal.api import array_autovisualizer
from treescope._internal.api import autovisualize as autovisualize_lib
from treescope._internal.api import default_renderer

# pylint: disable=import-outside-toplevel
try:
  import IPython
except ImportError:
  IPython = None
else:
  import IPython.core.formatters
  import IPython.core.magic
  import IPython.display
# pylint: enable=import-outside-toplevel


default_magic_autovisualizer: context.ContextualValue[
    autovisualize_lib.Autovisualizer
] = context.ContextualValue(
    module=__name__,
    qualname="default_magic_autovisualizer",
    initial_value=array_autovisualizer.ArrayAutovisualizer(),
)


def _display_and_maybe_steal(
    value: Any,
    ignore_exceptions: bool,
    roundtrip_mode: bool,
    streaming: bool,
    compress_html: bool,
    stealable: bool,
) -> str | None:
  """Helper to display a value, possibly with streaming and output stealing.

  Args:
    value: Value to display.
    ignore_exceptions: Whether to ignore exceptions during rendering.
    roundtrip_mode: Whether to start in roundtrip mode.
    streaming: Whether to render in streaming mode.
    compress_html: Whether to compress the HTML.
    stealable: Whether to return an HTML snippet that can be used to steal the
      output if it is already displayed.

  Returns:
    If `stealable` is True, returns an HTML snippet that can be used to steal
    the output if it is already displayed. Otherwise, returns None.
  """
  assert IPython is not None
  with contextlib.ExitStack() as stack:
    if streaming:
      deferreds = stack.enter_context(lowering.collecting_deferred_renderings())
    else:
      deferreds = []

    root_repr_method = object_inspection.safely_get_real_method(
        value, "__treescope_root_repr__"
    )
    if root_repr_method:
      foldable_ir = root_repr_method()
    else:
      foldable_ir = rendering_parts.build_full_line_with_annotations(
          default_renderer.build_foldable_representation(
              value, ignore_exceptions=ignore_exceptions
          )
      )
    if streaming:
      return lowering.display_streaming_as_root(
          foldable_ir,
          deferreds,
          roundtrip=roundtrip_mode,
          compressed=compress_html,
          stealable=stealable,
          ignore_exceptions=ignore_exceptions,
      )
    else:
      rendering = lowering.render_to_html_as_root(
          foldable_ir,
          roundtrip=roundtrip_mode,
          compressed=compress_html,
      )
      if stealable:
        return rendering
      else:
        IPython.display.display(IPython.display.HTML(rendering))


def display(
    value: Any,
    ignore_exceptions: bool = False,
    roundtrip_mode: bool = False,
    autovisualize: bool | autovisualize_lib.Autovisualizer | None = None,
    streaming: bool = True,
    compress_html: bool = True,
):
  """Displays a value as an interactively foldable object.

  Uses the default renderer.

  Args:
    value: Value to display.
    ignore_exceptions: Whether to catch errors during rendering of subtrees and
      show a fallback for those subtrees.
    roundtrip_mode: Whether to start in roundtrip mode.
    autovisualize: Optional autovisualizer override. If True, renders using the
      default autovisualizer (usually an array autovisualizer). If False,
      disables automatic visualization. If a function or object, uses that
      autovisualizer. If None (the default), uses the current active
      autovisualizer (if any) without overriding it.
    streaming: Whether to render in streaming mode, which immediately displays
      the structure of the output while computing more expensive leaf
      renderings. This is useful in interactive contexts, but can mess with
      other users of IPython's formatting because the final rendered HTML is
      empty.
    compress_html: Whether to zlib-compress (i.e. zip) treescope renderings to
      reduce their size when transmitted to the browser or saved into a
      notebook.

  Raises:
    RuntimeError: If IPython is not available.
  """
  if IPython is None:
    raise RuntimeError("Cannot use `display` outside of IPython.")
  with contextlib.ExitStack() as stack:
    if autovisualize is not None:
      if autovisualize is True:  # pylint: disable=g-bool-id-comparison
        tmp_autovisualizer = default_magic_autovisualizer.get()
      elif autovisualize is False:  # pylint: disable=g-bool-id-comparison
        tmp_autovisualizer = None
      else:
        tmp_autovisualizer = autovisualize
      stack.enter_context(
          autovisualize_lib.active_autovisualizer.set_scoped(tmp_autovisualizer)
      )
    maybe_stolen = _display_and_maybe_steal(
        value=value,
        ignore_exceptions=ignore_exceptions,
        roundtrip_mode=roundtrip_mode,
        streaming=streaming,
        compress_html=compress_html,
        stealable=False,
    )
    # Should not get an output when `stealable=False`
    assert maybe_stolen is None


def show(
    *args,
    wrap: bool = False,
    space_separated: bool = True,
    autovisualize: bool | autovisualize_lib.Autovisualizer | None = None,
):
  """Shows a list of objects inline, like python print, but with rich display.

  Args:
    *args: Values to show. Strings show as themselves, like python's print.
      Anything renderable by Treescope will show as it's treescope
      representation. Anything with a rich IPython representation will show as
      its IPython representation.
    wrap: Whether to wrap at the end of the line.
    space_separated: Whether to add single spaces between objects.
    autovisualize: Optional autovisualizer override. If True, renders using the
      default autovisualizer (usually an array autovisualizer). If False,
      disables automatic visualization. If a function or object, uses that
      autovisualizer. If None (the default), uses the current active
      autovisualizer (if any) without overriding it.

  Raises:
    RuntimeError: If IPython is not available.
  """
  if IPython is None:
    raise RuntimeError("Cannot use `show` outside of IPython.")
  if space_separated and args:
    separated_args = []
    for arg in args:
      separated_args.append(arg)
      separated_args.append(" ")
    args = separated_args[:-1]
  with contextlib.ExitStack() as stack:
    if autovisualize is not None:
      if autovisualize is True:  # pylint: disable=g-bool-id-comparison
        tmp_autovisualizer = default_magic_autovisualizer.get()
      elif autovisualize is False:  # pylint: disable=g-bool-id-comparison
        tmp_autovisualizer = None
      else:
        tmp_autovisualizer = autovisualize
      stack.enter_context(
          autovisualize_lib.active_autovisualizer.set_scoped(tmp_autovisualizer)
      )
    IPython.display.display(figures.inline(*args, wrap=wrap))


def register_as_default(
    streaming: bool = True,
    compress_html: bool = True,
):
  """Registers treescope as the default IPython renderer.

  This tells IPython to use treescope as a renderer for any object
  that doesn't have a specific renderer registered with IPython directly.

  Treescope will be configured to produce an HTML representation of most
  objects that do not have their own custom renderers. It will also be
  configured to produce summaries of jax.Array in text mode. Note that due to
  the way that IPython's text prettyprinter works, we can't easily set it up
  as a fallback renderer in text mode because IPython will prefer to use
  ordinary `repr` if it exists.

  Note that this hooks into every use of ``IPython.display.display(...)``. To
  avoid messing up ordinary display objects, if the object has a _repr_html_
  method already, we defer to that. (But if it's a structure containing display
  objects, we still use treescope as normal.)

  If the root object being rendered defines the special method
  `__treescope_root_repr__`, that method will be assumed to take no arguments
  and return a representation of the root object in Treescope's intermediate
  representation. This can be used to fully customize the rendering of a
  particular type of object. (Most types should instead define
  `__treescope_repr__`, which allows the rendering to be customized at any level
  of the tree, not just the root.)

  Args:
    streaming: Whether to render in streaming mode, which immediately displays
      the structure of the output while computing more expensive leaf
      renderings. This is useful in interactive contexts, but can mess with
      other users of IPython's formatting because the final rendered HTML is
      empty.
    compress_html: Whether to zlib-compress (i.e. zip) treescope renderings to
      reduce their size when transmitted to the browser or saved into a
      notebook.

  Raises:
    RuntimeError: If IPython is not available.
  """
  if IPython is None:
    raise RuntimeError("Cannot use `register_as_default` outside of IPython.")

  ipython_display = IPython.display

  def _render_for_ipython(value):
    repr_html_method = object_inspection.safely_get_real_method(
        value, "_repr_html_"
    )
    if repr_html_method:
      # Directly call the _repr_html_ method. We need to do this because if this
      # formatter returns None, the _repr_html_ method won't be called. (The
      # _repr_html_ method is only called if no formatter for the type is
      # found, but this formatter will run for every type due to being
      # registered for `object`.)
      return repr_html_method()  # pylint: disable=protected-access

    if isinstance(value, ipython_display.DisplayObject):
      # Don't render this to HTML, since IPython already knows how to render it.
      return None

    has_ipython_format_method = any(
        object_inspection.safely_get_real_method(value, method_name)
        for method_name in ("_repr_pretty_", "_repr_mimebundle_")
    )
    has_treescope_format_method = any(
        object_inspection.safely_get_real_method(value, method_name)
        for method_name in ("__treescope_repr__", "__treescope_root_repr__")
    )
    if has_ipython_format_method and not has_treescope_format_method:
      # Don't render this with Treescope, because it already has a different
      # rendering method for a non-HTML format.
      return None

    output_stealer = _display_and_maybe_steal(
        value=value,
        ignore_exceptions=True,
        roundtrip_mode=False,
        streaming=streaming,
        compress_html=compress_html,
        stealable=True,
    )
    # Executing the above call will have already displayed the output,
    # but it may be in the wrong place (e.g. it may appear before the
    # actual "Out" marker in JupyterLab). In `stealable` mode, by returning
    # `output_stealer` as the rendering of the object, we can ensure that the
    # output is moved to the right place.
    assert output_stealer is not None
    return output_stealer

  display_formatter = IPython.get_ipython().display_formatter
  cur_html_formatter = display_formatter.formatters["text/html"]
  cur_html_formatter.for_type(object, _render_for_ipython)

  def _render_as_text_oneline(value, p, cycle):
    del cycle
    with default_renderer.using_expansion_strategy(max_height=None):
      rendering = default_renderer.render_to_text(value, ignore_exceptions=True)
    for i, line in enumerate(rendering.split("\n")):
      if i:
        p.break_()
      p.text(line)

  # Override the text formatter to render jax.Array without copying the entire
  # array.
  cur_text_formatter = display_formatter.formatters["text/plain"]
  cur_text_formatter.for_type_by_name(
      "jaxlib.xla_extension", "ArrayImpl", _render_as_text_oneline
  )

  # Make sure the HTML formatter runs first, so streaming outputs work
  # correctly.
  old_formatters = display_formatter.formatters
  display_formatter.formatters = {}
  display_formatter.formatters["text/html"] = cur_html_formatter
  for k, v in old_formatters.items():
    if k != "text/html":
      display_formatter.formatters[k] = v

  try:
    from google.colab import _reprs  # pytype: disable=import-error  # pylint: disable=import-outside-toplevel,line-too-long

    _reprs.disable_string_repr()
    try:
      _reprs.disable_ndarray_repr()
    except KeyError:
      pass
    try:
      _reprs.disable_function_repr()
    except KeyError:
      pass
  except ImportError:
    pass


if IPython is not None:

  @IPython.core.magic.magics_class
  class AutovisualizerMagic(IPython.core.magic.Magics):
    """Magics class for enabling automatic visualization."""

    @IPython.core.magic.cell_magic
    def autovisualize(self, line, cell):
      """``%%autovisualize`` cell magic: enables autovisualization in a cell.

      The ``%%autovisualize`` magic is syntactic sugar for running a cell with
      automatic visualization of arrays (or a different automatic visualizer).
      This affects calls to `treescope.show` and `treescope.display` in the
      cell, and also affects the output formatting if Treescope is set as the
      default IPython pretty-printer.

      (Note that the ``%%autovisualize`` magic only affects cell output
      rendering if Treescope is already set as the default pretty-printer.)

      Allowed options:

      * ``%%autovisualize True`` or ``%%autovisualize`` with no arguments
        activates the default autovisualizer
        (`treescope.default_magic_autovisualizer`), which is usually an array
        autovisualizer.
      * ``%%autovisualize False`` or ``%%autovisualize None`` disables
        autovisualization in this cell.
      * ``%%autovisualize <autovisualizer_object>`` activates the given
        autovisualizer object.

      For instance, to use the default autovisualizer, you can annotate your
      cell with ::

        %%autovisualize

        # ... contents of your cell ...
        result = ...
        result

      which expands to something like ::

        with treescope.active_autovisualizer.set_scoped(
            treescope.default_magic_autovisualizer.get()
        ):
          # ... contents of your cell ...
          result = ...
          IPython.display.display(result)


      To activate an explicit autovisualizer function/object::

        %%autovisualize my_autovisualizer

        # ... contents of your cell ...
        result = ...
        result

      which expands to something like ::

        with treescope.active_autovisualizer.set_scoped(my_autovisualizer):
          # ... contents of your cell ...
          result = ...
          IPython.display.display(result)

      Args:
        line: Contents of the ``%%autovisualize`` magic, which will be evaluated
          as a Python object. If empty or `True`, the default autovisualizer
          will be used. If `False` or `None`, autovisualization will be
          disabled. Otherwise, the line should be a Python expression that
          evaluates to an autovisualizer object.
        cell: Contents of the rest of the cell. Will be run inside the
          autovisualization scope.
      """
      if line.strip():
        # Evaluate the line as Python code.
        autovisualizer = self.shell.ev(line)
      else:
        autovisualizer = True

      if autovisualizer is True:  # pylint: disable=g-bool-id-comparison
        # Retrieve the default autovisualizer.
        autovisualizer = default_magic_autovisualizer.get()
      elif autovisualizer is False:  # pylint: disable=g-bool-id-comparison
        # Disable autovisualization.
        autovisualizer = None

      with autovisualize_lib.active_autovisualizer.set_scoped(autovisualizer):
        self.shell.run_cell(cell)

  @IPython.core.magic.magics_class
  class ContextManagerMagic(IPython.core.magic.Magics):
    """Magics class for using ``%%with`` to run a cell in a context manager."""

    @IPython.core.magic.cell_magic("with")
    def with_(self, line, cell):
      """``%%with`` cell magic: runs under a context manager.

      The ``%%with`` magic is syntactic sugar for running a cell under a context
      manager. It can be used to easily set Treescope's context variables in a
      cell without having to explicitly display the final output. In other
      words, ::

        %%with some_var.set_scoped(Foo)
        result = ...
        result

      expands to something like ::

        with some_var.set_scoped(Foo):
          result = ...
          IPython.display.display(result)

      Args:
        line: Contents of the line where ``%%with`` is. Should be a Python
          expression for a context manager.
        cell: Contents of the rest of the cell. Will be run inside the context
          manager scope.
      """
      ctxmgr = self.shell.ev(line)
      with ctxmgr:
        self.shell.run_cell(cell)

else:
  AutovisualizerMagic = None  # pylint: disable=invalid-name
  ContextManagerMagic = None  # pylint: disable=invalid-name


def register_autovisualize_magic():
  """Registers the ``%%autovisualize`` magic.

  This makes it possible to use ``%%autovisualize`` at the top of a cell to
  enable automatic visualization for that cell's outputs.

  Raises:
    RuntimeError: If IPython is not available.
  """
  if IPython is None:
    raise RuntimeError(
        "Cannot use `register_autovisualize_magic` outside of IPython."
    )
  IPython.get_ipython().register_magics(AutovisualizerMagic)


def register_context_manager_magic():
  """Registers the ``%%with`` magic.

  This makes it possible to use ``%%with`` at the top of a cell to enable
  automatic visualization for that cell's outputs.

  Raises:
    RuntimeError: If IPython is not available.
  """
  if IPython is None:
    raise RuntimeError(
        "Cannot use `register_context_manager_magic` outside of IPython."
    )
  IPython.get_ipython().register_magics(ContextManagerMagic)


def basic_interactive_setup(autovisualize_arrays: bool = True):
  """Sets up IPython for interactive use with Treescope.

  This is a helper function that runs various setup steps:

    * Configures Treescope as the default IPython renderer.
    * Turns on interactive mode for Treescope's context managers.
    * Registers the `%%autovisualize` magic.
    * Registers the `%%with` magic.
    * If `autovisualize_arrays` is True, configures Treescope to automatically
      visualize arrays.

  Args:
    autovisualize_arrays: Whether to automatically visualize arrays.
  """
  register_as_default()
  register_autovisualize_magic()
  register_context_manager_magic()

  if autovisualize_arrays:
    autovisualize_lib.active_autovisualizer.set_globally(
        array_autovisualizer.ArrayAutovisualizer()
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler for custom types via __treescope_repr__ or the global registry."""

from __future__ import annotations

from typing import Any

from treescope import renderers
from treescope import rendering_parts
from treescope import type_registries
from treescope._internal import object_inspection


def handle_via_treescope_repr_method(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
    *,
    method_name: str = "__treescope_repr__",
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a type by calling its ``__treescope_repr__`` method, if it exists.

  The ``__treescope_repr__`` method can be used to add treescope support to
  custom classes. The method is expected to return a rendering in treescope's
  internal intermediate representation.

  The exact structure of the intermediate representation is an
  implementation detail and may change in future releases. Users should instead
  build a rendering using the high-level helpers in `treescope.repr_lib`, or
  the lower-level definitions in `treescope.rendering_parts`.

  Args:
    node: The node to render.
    path: An optional path to this node from the root.
    subtree_renderer: The renderer for subtrees of this node.
    method_name: Optional override for the name of the method to call. This is
      primarily intended for compatibility with treescope's original
      implementation as part of Penzai, which used __penzai_repr__.

  Returns:
    A rendering of this node, if it implements the __treescope_repr__ extension
    method.
  """
  treescope_repr_method = object_inspection.safely_get_real_method(
      node, method_name
  )
  if treescope_repr_method:
    return treescope_repr_method(path, subtree_renderer)
  else:
    return NotImplemented


def handle_via_global_registry(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a type by looking it up in the global handler registry.

  If it is not feasible to define ``__treescope_repr__`` for a type, it can
  instead be registered in the global handler registry. This is a dictionary
  mapping types to functions that render a node of that type.

  The exact structure of the intermediate representation is an
  implementation detail and may change in future releases. Users should instead
  build a rendering using the high-level helpers in `treescope.repr_lib`, or
  the lower-level definitions in `treescope.rendering_parts`.

  Args:
    node: The node to render.
    path: An optional path to this node from the root.
    subtree_renderer: The renderer for subtrees of this node.

  Returns:
    A rendering of this node, if it was found in the global registry.
  """
  maybe_handler = type_registries.lookup_treescope_handler_for_type(type(node))
  if maybe_handler:
    return maybe_handler(node, path, subtree_renderer)
  else:
    return NotImplemented



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wraps MyST-nb output cells in HTML templates.

This can be used to sandbox the environment of output cell renderings, which is
useful for Treescope renderings since they make heavy use of JavaScript and CSS,
and can also improve responsiveness of the main page.

This transformation replaces output cells with ``<template>`` elements tagged
with the class ``cell_output_frame_src``. This can then be detected by the
``custom.js`` javascript, which will construct ``<iframe>`` elements with those
contents.
"""

from typing import Any, Mapping

from docutils import nodes
from sphinx import application
from sphinx.transforms import post_transforms


class OutputCellsToIframe(post_transforms.SphinxPostTransform):
  """Wraps output cell renderings in HTML templates."""

  default_priority = 199
  formats = ("html",)

  def run(self, **kwargs):
    for node in self.document.findall(nodes.container):
      if (
          "glued-cell-output" in node["classes"]
          or node.get("nb_element") == "cell_code_output"
      ):
        node.insert(
            0,
            nodes.raw(
                text='<template class="cell_output_frame_src">', format="html"
            ),
        )
        node.append(nodes.raw(text="</template>", format="html"))


def setup(app: application.Sphinx) -> Mapping[str, Any]:
  app.add_post_transform(OutputCellsToIframe)
  return dict(version="0.1", parallel_read_safe=True)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler that falls back to ordinary `repr`."""

from __future__ import annotations

from typing import Any

from treescope import renderers
from treescope import rendering_parts


def handle_anything_with_repr(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Builds a foldable from its repr."""
  del subtree_renderer
  node_repr = repr(node)
  basic_repr = object.__repr__(node)
  if "\n" in node_repr:
    # Multiline repr. Use a simpler version for the one-line summary.
    if node_repr.startswith("<") and node_repr.endswith(">"):
      # This node has an idiomatic non-roundtrippable repr.
      # Replace its newlines with markers, and add the basic summary as a
      # comment
      lines = node_repr.split("\n")
      lines_with_markers = [
          rendering_parts.siblings(
              rendering_parts.text(line),
              rendering_parts.fold_condition(
                  expanded=rendering_parts.text("\n"),
                  collapsed=rendering_parts.comment_color(
                      rendering_parts.text("↩")
                  ),
              ),
          )
          for line in lines[:-1]
      ]
      lines_with_markers.append(rendering_parts.text(lines[-1]))
      return rendering_parts.siblings_with_annotations(
          rendering_parts.build_custom_foldable_tree_node(
              contents=rendering_parts.abbreviation_color(
                  rendering_parts.siblings(*lines_with_markers)
              ),
              path=path,
          ),
          extra_annotations=[
              rendering_parts.comment_color(
                  rendering_parts.text("  # " + basic_repr)
              )
          ],
      )
    else:
      # Use basic repr as the summary.
      return rendering_parts.build_custom_foldable_tree_node(
          label=rendering_parts.abbreviation_color(
              rendering_parts.comment_color_when_expanded(
                  rendering_parts.siblings(
                      rendering_parts.fold_condition(
                          expanded=rendering_parts.text("# ")
                      ),
                      basic_repr,
                  )
              )
          ),
          contents=rendering_parts.fold_condition(
              expanded=rendering_parts.indented_children([
                  rendering_parts.abbreviation_color(
                      rendering_parts.text(node_repr)
                  )
              ])
          ),
          path=path,
      )
  elif node_repr == basic_repr:
    # Just use the basic repr as the summary, since we don't have anything else.
    return rendering_parts.build_one_line_tree_node(
        line=rendering_parts.abbreviation_color(
            rendering_parts.text(node_repr)
        ),
        path=path,
    )
  else:
    # Use the custom repr as a one-line summary, but float the basic repr to
    # the right to tell the user what the type is in case the custom repr
    # doesn't include that info.
    return rendering_parts.siblings_with_annotations(
        rendering_parts.build_one_line_tree_node(
            line=rendering_parts.abbreviation_color(
                rendering_parts.text(node_repr)
            ),
            path=path,
        ),
        extra_annotations=[
            rendering_parts.comment_color(
                rendering_parts.text("  # " + basic_repr)
            )
        ],
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for formatting and rendering."""

import hashlib
import warnings


def oklch_color(
    lightness: float, chroma: float, hue: float, alpha: float = 1.0
) -> str:
  """Constructs an OKLCH CSS color.

  See https://developer.mozilla.org/en-US/docs/Web/CSS/color_value/oklch and
  https://oklch.com/.

  Args:
    lightness: Lightness argument (0 to 1)
    chroma: Chroma argument (0 to 0.5 in practice; large values are clipped)
    hue: Hue argument (0 to 360)
    alpha: Opacity (0 to 1)

  Returns:
    A CSS color string.
  """
  return f"oklch({lightness} {chroma} {hue} / {alpha})"


def color_from_string(
    key_string: str,
    lightness: float = 0.785,
    chroma: float = 0.103,
    alpha: float = 1.0,
) -> str:
  """Derives a color whose hue is keyed by a string.

  The main use for this is to automatically generate a color for the CSS OKLCH
  color space, which can be used to set a type-dependent color for an object in
  treescope.

  The default values of ``lightness`` and ``chroma`` are set up to give a
  consistent range of colors that are all within gamut.

  Args:
    key_string: A string that should be hashed to determine a hue.
    lightness: Lightness argument (0 to 1)
    chroma: Chroma argument (0 to 0.5 in practice; large values are clipped)
    alpha: Opacity (0 to 1)

  Returns:
    A pseudo-uniformly-distributed number between 0 and 360, suitable for use
    as the third argument to the CSS ``oklch(...)`` function.
  """
  # Derive a 32-bit fingerprint from the type.
  fingerprint = int.from_bytes(
      hashlib.sha256(key_string.encode("utf-8")).digest(), byteorder="little"
  )
  # fingerprint = farmhash.fingerprint32(key_string)
  # Convert it into a uniformly-distributed float in [0, 1].
  uniform = (fingerprint % (2**16)) / (2**16)
  # Convert to a hue.
  return oklch_color(lightness, chroma, 360 * uniform, alpha)


def parse_simple_color_and_pattern_spec(
    requested_color: str | tuple[str, str],
    typename_for_warning: str | None = None,
) -> tuple[str | None, str | None]:
  """Parses a background color and pattern from a user-provided color request.

  Args:
    requested_color: A color request, which is either a single CSS color or a
      tuple of outline and background colors.
    typename_for_warning: If provided, and the color is invalid, a warning will
      be issued with this typename as context.

  Returns:
    A tuple (background_color, background_pattern) that can be passed to
    the Treescope low-level representation construction functions (such as
    `build_foldable_tree_node_from_children`) that will configure it with the
    given background color and outline.
  """
  if isinstance(requested_color, str):
    background_color = requested_color
    background_pattern = None
  elif isinstance(requested_color, tuple) and len(requested_color) == 2:
    background_color = requested_color[0]
    background_pattern = (
        f"linear-gradient({requested_color[1]},{requested_color[1]})"
    )
  else:
    if typename_for_warning:
      warnings.warn(
          f"{typename_for_warning} requested an invalid color:"
          f" {requested_color} (not a string or a tuple)"
      )
    background_color = None
    background_pattern = None
  return background_color, background_pattern



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for documentation generation."""

from typing import Any

_SKIPPED_FOR_AUTODOC = {}


def skip_automatic_documentation(obj: Any):
  """Marks an object as skipped for automatic documentation generation."""
  _SKIPPED_FOR_AUTODOC[id(obj)] = obj
  return obj



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common building blocks for the treescope intermediate representation."""

from __future__ import annotations

import abc
from collections.abc import Hashable
import dataclasses
import functools
import html
import io
import itertools
import operator
import typing
from typing import Any, Sequence

from treescope._internal import html_escaping
from treescope._internal.parts import part_interface


CSSStyleRule = part_interface.CSSStyleRule
JavaScriptDefn = part_interface.JavaScriptDefn
HtmlContextForSetup = part_interface.HtmlContextForSetup
RenderableTreePart = part_interface.RenderableTreePart
ExpandState = part_interface.ExpandState
FoldableTreeNode = part_interface.FoldableTreeNode
RenderableAndLineAnnotations = part_interface.RenderableAndLineAnnotations

################################################################################
# Basic combinators
################################################################################


@dataclasses.dataclass(frozen=True)
class BaseContentlessLeaf(RenderableTreePart):
  """A part whose defaults indicate no renderable content, for subclassing."""

  def _compute_collapsed_width(self) -> int:
    return 0

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 0

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    return set()

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    pass

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    pass


@dataclasses.dataclass(frozen=True)
@typing.final
class EmptyPart(BaseContentlessLeaf):
  """A definitely-empty part, which can be detected and special-cased."""


def empty_part() -> EmptyPart:
  """Returns an empty part."""
  return EmptyPart()


@dataclasses.dataclass(frozen=True)
class Text(RenderableTreePart):
  """A raw text literal."""

  text: str

  def _compute_collapsed_width(self) -> int:
    if "\n" in self.text:
      raise ValueError(
          "Cannot compute collapsed length for a text literal with a newline!"
          " Text literals with newlines need to be wrapped to only show in"
          " expanded mode."
      )
    return len(self.text)

  def _compute_newlines_in_expanded_parent(self) -> int:
    return self.text.count("\n")

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    return set()

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write(html.escape(self.text))

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if "\n" in self.text and not expanded_parent:
      raise ValueError(
          "Cannot render a text literal with a newline in collapsed mode!"
          " Text literals with newlines need to be wrapped to only show in"
          " expanded mode."
      )
    stream.write(("\n" + " " * indent).join(self.text.split("\n")))


def text(text_content: str) -> RenderableTreePart:
  """Builds a one-line text part."""
  return Text(text_content)


@dataclasses.dataclass(frozen=True)
class Siblings(RenderableTreePart):
  """A sequence of children parts, rendered inline."""

  children: Sequence[RenderableTreePart]

  def _compute_collapsed_width(self) -> int:
    return sum(part.collapsed_width for part in self.children)

  def _compute_newlines_in_expanded_parent(self) -> int:
    return sum(part.newlines_in_expanded_parent for part in self.children)

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return list(
        itertools.chain(
            *(part.foldables_in_this_part() for part in self.children)
        )
    )

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return functools.reduce(
        operator.or_,
        (part.layout_marks_in_this_part for part in self.children),
        frozenset(),
    )

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    return functools.reduce(
        operator.or_,
        (part.html_setup_parts(context) for part in self.children),
        set(),
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    for part in self.children:
      part.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      # Only the first part gets beginning-of-line privileges.
      at_beginning_of_line = False

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    for part in self.children:
      part.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


def siblings(*args: RenderableTreePart | str) -> RenderableTreePart:
  """Builds a Siblings part from inline arguments.

  Args:
    *args: Sequence of renderables or strings (which will be wrapped in Text).

  Returns:
    A new Siblings part containing these concatenated together.
  """
  parts = []
  for arg in args:
    if isinstance(arg, str):
      parts.append(Text(arg))
    elif isinstance(arg, Siblings):
      parts.extend(arg.children)
    elif isinstance(arg, EmptyPart):
      pass
    elif isinstance(arg, RenderableTreePart):
      parts.append(arg)
    else:
      raise ValueError(f"Invalid argument type {type(arg)}")
  return Siblings(tuple(parts))


class DeferringToChild(RenderableTreePart):
  """Helper base class that defers methods to a child attribute.

  Subclasses can derive from this to get default implementations for all
  required methods, then override the ones where they have custom behavior.
  """

  # note: not a dataclass, just here for type annotations
  child: RenderableTreePart

  def _compute_collapsed_width(self) -> int:
    return self.child.collapsed_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    return self.child.newlines_in_expanded_parent

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return self.child.foldables_in_this_part()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return self.child.layout_marks_in_this_part

  def html_setup_parts(
      self, context: HtmlContextForSetup, /
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    return self.child.html_setup_parts(context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    self.child.render_to_text(
        stream,
        expanded_parent=expanded_parent,
        indent=indent,
        roundtrip_mode=roundtrip_mode,
        render_context=render_context,
    )


@dataclasses.dataclass(frozen=True)
class BaseSpanGroup(DeferringToChild, abc.ABC):
  """A group that wraps its child in a span, and may apply CSS styles.

  Subclasses can define a particular CSS style to apply.

  Attributes:
    child: Contents of the group.
  """

  child: RenderableTreePart

  # Overridable methods defined here.

  @abc.abstractmethod
  def _span_css_class(self) -> str:
    """Returns a CSS class for the span's style. Intended to be overridden."""
    raise NotImplementedError("_span_css_class should be overridden")

  @abc.abstractmethod
  def _span_css_rule(
      self, context: HtmlContextForSetup, /
  ) -> CSSStyleRule | set[CSSStyleRule]:
    """Returns a CSS style rule or rules for the class in `_span_css_class`.

    Intended to be overridden.

    Args:
      context: Context for setting up.
    """
    raise NotImplementedError("_span_css_rule should be overridden")

  # Implementations of parent class abstract methods.

  def html_setup_parts(
      self, context: HtmlContextForSetup, /
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    my_rule = self._span_css_rule(context)
    if isinstance(my_rule, CSSStyleRule):
      my_rule = {my_rule}
    return my_rule | self.child.html_setup_parts(context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    class_string = self._span_css_class()
    stream.write(f'<span class="{class_string}">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


@dataclasses.dataclass(frozen=True)
class WithLayoutMark(DeferringToChild):
  """A group that marks its child for layout purposes.

  Attributes:
    child: Contents of the group.
    mark: A layout mark to apply to the child. This can later be used by layout
      algorithms to show this node.
  """

  child: RenderableTreePart
  mark: Hashable

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset((self.mark,)) | self.child.layout_marks_in_this_part


def with_layout_mark(
    child: RenderableTreePart, mark: Hashable
) -> RenderableTreePart:
  """Returns a part that marks its child for layout purposes.

  Args:
    child: Contents of the group.
    mark: A layout mark to apply to the child. This can later be used by layout
      algorithms to show this node.

  Returns:
    A new part that marks its child for layout purposes.
  """
  if not isinstance(child, RenderableTreePart):
    raise ValueError(f"child must be a renderable part, got {type(child)}")
  if not isinstance(mark, Hashable):
    raise ValueError(f"Layout marks must be hashable, got {mark}")
  return WithLayoutMark(child=child, mark=mark)


@dataclasses.dataclass(frozen=True)
class VerticalSpace(RenderableTreePart):
  """A vertical space in HTML mode."""

  height: str

  def _compute_collapsed_width(self) -> int:
    return 0

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 0

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    return set()

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    pass

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write(
        "<div"
        f' style="height:{html_escaping.escape_html_attribute(self.height)};'
        ' width:0;"></div>'
    )


def vertical_space(css_height: str) -> RenderableTreePart:
  """Returns a vertical space with the given height in HTML mode.

  Args:
    css_height: The height of the space, as a CSS length string.

  Returns:
    A renderable part that renders as a vertical space in HTML mode, and does
    not render in text mode.
  """
  if not isinstance(css_height, str):
    raise ValueError(f"css_height must be a string, got {css_height}")
  return VerticalSpace(height=css_height)


################################################################################
# Conditional rendering
################################################################################


@dataclasses.dataclass(frozen=True)
class FoldCondition(RenderableTreePart):
  """Renders conditionally depending on whether it's collapsed or expanded.

  Attributes:
    collapsed: Contents to render when parent is collapsed.
    expanded: Contents to render when parent is expanded.
  """

  collapsed: RenderableTreePart = EmptyPart()
  expanded: RenderableTreePart = EmptyPart()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .when_collapsed {{
            display: none;
        }}
        {context.collapsed_selector} .when_collapsed {{
            display: inline;
        }}
        .when_expanded {{
            display: inline;
        }}
        {context.collapsed_selector} .when_expanded {{
            display: none;
        }}
        """)
    return (
        {CSSStyleRule(rule)}
        | self.collapsed.html_setup_parts(context)
        | self.expanded.html_setup_parts(context)
    )

  def _compute_collapsed_width(self) -> int:
    return self.collapsed.collapsed_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    # Disappears when expanded.
    return self.expanded.newlines_in_expanded_parent

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    # Only the part that's visible when expanded will have foldables that are
    # possible to fold/unfold.
    return self.expanded.foldables_in_this_part()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return (
        self.collapsed.layout_marks_in_this_part
        | self.expanded.layout_marks_in_this_part
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if not isinstance(self.collapsed, EmptyPart):
      stream.write('<span class="when_collapsed">')
      self.collapsed.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")
    if not isinstance(self.expanded, EmptyPart):
      stream.write('<span class="when_expanded">')
      self.expanded.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if expanded_parent:
      self.expanded.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )
    else:
      self.collapsed.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


def fold_condition(
    collapsed: RenderableTreePart | None = None,
    expanded: RenderableTreePart | None = None,
) -> RenderableTreePart:
  """Builds a part that renders differently when collapsed or expanded.

  Args:
    collapsed: Contents to render when parent is collapsed.
    expanded: Contents to render when parent is expanded.

  Returns:
    A renderable part that renders as ``collapsed`` when the parent is collapsed
    and as ``expanded`` when the parent is expanded.
  """
  if collapsed is None:
    collapsed = EmptyPart()
  if expanded is None:
    expanded = EmptyPart()
  if not isinstance(collapsed, RenderableTreePart):
    raise ValueError(
        "`collapsed` must be a renderable part or None. Got"
        f" {type(collapsed).__name__}"
    )
  if not isinstance(expanded, RenderableTreePart):
    raise ValueError(
        "`expanded` must be a renderable part or None. Got"
        f" {type(expanded).__name__}"
    )
  return FoldCondition(collapsed=collapsed, expanded=expanded)


@dataclasses.dataclass(frozen=True)
class RoundtripCondition(RenderableTreePart):
  """Renders conditionally depending on whether it's in roundtrip mode.

  Attributes:
    roundtrip: Contents to render when rendering in round trip mode.
    not_roundtrip: Contents to render when rendering in ordinary mode.
  """

  roundtrip: RenderableTreePart = EmptyPart()
  not_roundtrip: RenderableTreePart = EmptyPart()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .when_roundtrip {{
            display: none;
        }}
        {context.roundtrip_selector} .when_roundtrip {{
            display: inline;
        }}
        .when_not_roundtrip {{
            display: inline;
        }}
        {context.roundtrip_selector} .when_not_roundtrip {{
            display: none;
        }}
        """)
    return (
        {CSSStyleRule(rule)}
        | self.roundtrip.html_setup_parts(context)
        | self.not_roundtrip.html_setup_parts(context)
    )

  def _compute_collapsed_width(self) -> int:
    # Defer to the not_roundtrip version for expansion decisions.
    return self.not_roundtrip.collapsed_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    # Defer to the not_roundtrip version for expansion decisions.
    return self.not_roundtrip.newlines_in_expanded_parent

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return (
        self.roundtrip.layout_marks_in_this_part
        | self.not_roundtrip.layout_marks_in_this_part
    )

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    # Defer to the not_roundtrip version for expansion decisions.
    return self.not_roundtrip.foldables_in_this_part()

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if not isinstance(self.roundtrip, EmptyPart):
      stream.write('<span class="when_roundtrip">')
      self.roundtrip.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")

    if not isinstance(self.not_roundtrip, EmptyPart):
      stream.write('<span class="when_not_roundtrip">')
      self.not_roundtrip.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if roundtrip_mode:
      self.roundtrip.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )
    else:
      self.not_roundtrip.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


def roundtrip_condition(
    roundtrip: RenderableTreePart | None = None,
    not_roundtrip: RenderableTreePart | None = None,
) -> RenderableTreePart:
  """Builds a part that renders differently in roundtrip mode.

  Args:
    roundtrip: Contents to render when rendering in round trip mode.
    not_roundtrip: Contents to render when rendering in ordinary mode.

  Returns:
    A renderable part that renders as ``roundtrip`` in roundtrip mode
    and as ``not_roundtrip`` in ordinary mode.
  """
  if roundtrip is None:
    roundtrip = EmptyPart()
  if not_roundtrip is None:
    not_roundtrip = EmptyPart()
  if not isinstance(roundtrip, RenderableTreePart):
    raise ValueError(
        "`roundtrip` must be a renderable part or None. Got"
        f" {type(roundtrip).__name__}"
    )
  if not isinstance(not_roundtrip, RenderableTreePart):
    raise ValueError(
        "`not_roundtrip` must be a renderable part or None. Got"
        f" {type(not_roundtrip).__name__}"
    )
  return RoundtripCondition(roundtrip=roundtrip, not_roundtrip=not_roundtrip)


@dataclasses.dataclass(frozen=True)
class SummarizableCondition(RenderableTreePart):
  """Renders conditionally depending on combination of roundtrip/collapsed.

  The idea is that, when collapsed and not in roundtrip mode, it's sometimes
  convenient to summarize a compound node with a simpler non-roundtrippable
  representation.

  Attributes:
    summary: Contents to render when collapsed and not in roundtrip mode.
    detail: Contents to render when either expanded or in roundtrip mode.
  """

  summary: RenderableTreePart = EmptyPart()
  detail: RenderableTreePart = EmptyPart()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .when_expanded_or_roundtrip {{
            display: none;
        }}
        .when_expanded_or_roundtrip:not({context.collapsed_selector} *),
        {context.roundtrip_selector} .when_expanded_or_roundtrip {{
            display: inline;
        }}
        .when_collapsed_and_not_roundtrip {{
            display: inline;
        }}
        .when_collapsed_and_not_roundtrip:not({context.collapsed_selector} *),
        {context.roundtrip_selector} .when_collapsed_and_not_roundtrip {{
            display: none;
        }}
        """)
    return (
        {CSSStyleRule(rule)}
        | self.summary.html_setup_parts(context)
        | self.detail.html_setup_parts(context)
    )

  def _compute_collapsed_width(self) -> int:
    # Defer to the summary version for expansion decisions.
    return self.summary.collapsed_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    # Defer to the detail version for expansion decisions.
    return self.detail.newlines_in_expanded_parent

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return (
        self.summary.layout_marks_in_this_part
        | self.detail.layout_marks_in_this_part
    )

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    # Defer to the detail version for expansion decisions.
    return self.detail.foldables_in_this_part()

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if not isinstance(self.summary, EmptyPart):
      stream.write('<span class="when_collapsed_and_not_roundtrip">')
      self.summary.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")

    if not isinstance(self.detail, EmptyPart):
      stream.write('<span class="when_expanded_or_roundtrip">')
      self.detail.render_to_html(
          stream,
          at_beginning_of_line=at_beginning_of_line,
          render_context=render_context,
      )
      stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if not expanded_parent and not roundtrip_mode:
      self.summary.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )
    else:
      self.detail.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


def summarizable_condition(
    summary: RenderableTreePart | None = None,
    detail: RenderableTreePart | None = None,
) -> RenderableTreePart:
  """Builds a part that renders depending on combination of roundtrip/collapsed.

  The idea is that, when collapsed and not in roundtrip mode, it's sometimes
  convenient to summarize a compound node with a simpler non-roundtrippable
  representation.

  Args:
    summary: Contents to render when collapsed and not in roundtrip mode.
    detail: Contents to render when either expanded or in roundtrip mode.

  Returns:
    A renderable part that renders as ``summary`` when both collapsed and not
    in roundtrip mode, and as ``detail`` otherwise.
  """
  if summary is None:
    summary = EmptyPart()
  if detail is None:
    detail = EmptyPart()
  if not isinstance(summary, RenderableTreePart):
    raise ValueError(
        "`summary` must be a renderable part or None. Got"
        f" {type(summary).__name__}"
    )
  if not isinstance(detail, RenderableTreePart):
    raise ValueError(
        "`detail` must be a renderable part or None. Got"
        f" {type(detail).__name__}"
    )
  return SummarizableCondition(summary=summary, detail=detail)


################################################################################
# Line comments
################################################################################


def siblings_with_annotations(
    *args: str | RenderableTreePart | RenderableAndLineAnnotations,
    extra_annotations: Sequence[RenderableTreePart] = (),
) -> RenderableAndLineAnnotations:
  """Combines siblings that may have annotations, aggregating separately.

  This can be used to lay out multiple objects on the same line, when some
  may have annotations.

  Args:
    *args: Sequence of strings, renderable tree parts, or commented tree parts
      to render.
    extra_annotations: Additional annotations to add.

  Returns:
    A new pair of renderable and annotations, with main renderables and
    annotations combined separately.
  """
  parts = []
  annotations = []
  for arg in args:
    if isinstance(arg, RenderableTreePart):
      parts.append(arg)
    elif isinstance(arg, str):
      parts.append(Text(arg))
    elif isinstance(arg, RenderableAndLineAnnotations):
      parts.append(arg.renderable)
      if arg.annotations is not None:
        annotations.append(arg.annotations)
    else:
      raise ValueError(
          "Expected a renderable tree part (possibly with line annotations) or"
          f" a string, but got: {type(arg)}"
      )

  for annotation in extra_annotations:
    annotations.append(annotation)

  return RenderableAndLineAnnotations(siblings(*parts), siblings(*annotations))


def build_full_line_with_annotations(
    *args: str | RenderableTreePart | RenderableAndLineAnnotations,
) -> RenderableTreePart:
  """Concatenates tree parts and appends their line comments at the end.

  To ensure that the output is formatted correctly, the output of this function
  should always appear at the end of a line when its parent is expanded. This
  is not checked. (Failing to do this may cause delimiters to be erroneously
  commented out.)

  Args:
    *args: Sequence of strings, renderable tree parts, or commented tree parts
      to render.

  Returns:
    A renderable tree part that combines all of the main renderables, and also
    shows the line annotations on the right when expanded.
  """
  combined = siblings_with_annotations(*args)
  if (
      combined.annotations is None
      or isinstance(combined.annotations, EmptyPart)
      or (
          isinstance(combined.annotations, Siblings)
          and not combined.annotations.children
      )
  ):
    return combined.renderable
  return siblings(
      combined.renderable,
      FoldCondition(expanded=combined.annotations),
  )


################################################################################
# Structure combinators
################################################################################


@dataclasses.dataclass(frozen=True)
class OnSeparateLines(RenderableTreePart):
  """A sequence of children, one per line, not indented.

  Attributes:
    children: Children to render on separate lines when expanded.
  """

  children: Sequence[RenderableTreePart]

  @classmethod
  def build(
      cls,
      children: Sequence[RenderableAndLineAnnotations | RenderableTreePart],
  ) -> OnSeparateLines:
    """Builds a OnSeparateLines instance, supporting annotations."""
    return cls([build_full_line_with_annotations(line) for line in children])

  def _compute_collapsed_width(self) -> int:
    # When collapsed, all children appear inline.
    return sum(part.collapsed_width for part in self.children)

  def _compute_newlines_in_expanded_parent(self) -> int:
    # When expanded, we format every child on its own line. This introduces
    # len(children) + 1 newlines total, under the assumption that there was
    # some inline content before and after this IndentedChildren node.
    return (
        sum(part.newlines_in_expanded_parent + 1 for part in self.children) + 1
    )

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return functools.reduce(
        operator.or_,
        (part.layout_marks_in_this_part for part in self.children),
        frozenset(),
    )

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return list(
        itertools.chain(
            *(part.foldables_in_this_part() for part in self.children)
        )
    )

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .separate_lines_child:not({context.collapsed_selector} *)
        {{
            display: block;
        }}
        .separate_lines_children:not({context.collapsed_selector} *)
        {{
            display: block;
        }}
        """)
    return functools.reduce(
        operator.or_,
        (part.html_setup_parts(context) for part in self.children),
        {CSSStyleRule(rule)},
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write('<span class="separate_lines_children">')
    for part in self.children:
      # All children render at the beginning of their line.
      stream.write('<span class="separate_lines_child">')
      part.render_to_html(
          stream,
          at_beginning_of_line=True,
          render_context=render_context,
      )
      stream.write("</span>")

    stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # Insert separators before and after each child if in expanded mode.
    if expanded_parent:
      separator = "\n" + " " * indent
    else:
      separator = ""

    for line_part in self.children:
      stream.write(separator)
      line_part.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )

    if expanded_parent:
      stream.write("\n" + " " * indent)


def on_separate_lines(
    children: Sequence[RenderableAndLineAnnotations | RenderableTreePart],
) -> RenderableTreePart:
  """Builds a part that renders its children on separate lines.

  The resulting part stacks the children together, moving any comments to the
  end of their lines.

  Args:
    children: Children to render.

  Returns:
    A renderable part that renders the children on separate lines when expanded.
    When collapsed, it instead concatenates them.
  """
  return OnSeparateLines.build(children)


@dataclasses.dataclass(frozen=True)
class IndentedChildren(RenderableTreePart):
  """A sequence of children, one per line, and indented.

  IndentedChildren is the primary way to lay out the content of ordinary
  containers across multiple lines.

  Attributes:
    children: Children to render on separate lines when expanded.
  """

  children: Sequence[RenderableTreePart]

  @classmethod
  def build(
      cls,
      children: Sequence[RenderableAndLineAnnotations | RenderableTreePart],
      comma_separated: bool = False,
      force_trailing_comma: bool = False,
  ) -> IndentedChildren:
    """Builds a IndentedChildren instance."""
    lines = []
    for i, child in enumerate(children):
      if comma_separated:
        if i < len(children) - 1:
          # Not the last child. Always show a comma, and add a space when
          # collapsed.
          delimiter = siblings(",", FoldCondition(collapsed=Text(" ")))
        elif force_trailing_comma:
          # Last child, forced comma.
          delimiter = Text(",")
        else:
          # Last child, but only show the comma when the node is expanded.
          delimiter = FoldCondition(expanded=Text(","))
        line = build_full_line_with_annotations(child, delimiter)
      else:
        line = build_full_line_with_annotations(child)

      lines.append(line)

    return cls(lines)

  def _compute_collapsed_width(self) -> int:
    # When collapsed, all children appear inline.
    return sum(part.collapsed_width for part in self.children)

  def _compute_newlines_in_expanded_parent(self) -> int:
    # When expanded, we format every child on its own line. This introduces
    # len(children) + 1 newlines total, under the assumption that there was
    # some inline content before and after this IndentedChildren node.
    return (
        sum(part.newlines_in_expanded_parent + 1 for part in self.children) + 1
    )

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return functools.reduce(
        operator.or_,
        (part.layout_marks_in_this_part for part in self.children),
        frozenset(),
    )

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return list(
        itertools.chain(
            *(part.foldables_in_this_part() for part in self.children)
        )
    )

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .indented_children {{
            contain: content;
        }}
        .indented_child:not({context.collapsed_selector} *)
        {{
            display: block;
            margin-left: calc(2ch - 1px);
        }}
        .stacked_children:not({context.collapsed_selector} *)
        {{
            display: block;
            border-left: dotted 1px #e0e0e0;
        }}
        """)
    return functools.reduce(
        operator.or_,
        (part.html_setup_parts(context) for part in self.children),
        {CSSStyleRule(rule)},
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write('<span class="indented_children">')
    for part in self.children:
      # All children render at the beginning of their line.
      stream.write('<span class="indented_child">')
      part.render_to_html(
          stream,
          at_beginning_of_line=True,
          render_context=render_context,
      )
      stream.write("</span>")

    stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # Insert separators before and after each child if in expanded mode.
    if expanded_parent:
      next_indent = indent + 2
      separator = "\n" + " " * next_indent
    else:
      next_indent = indent
      separator = ""

    for line_part in self.children:
      stream.write(separator)
      line_part.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=next_indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )

    if expanded_parent:
      stream.write("\n" + " " * indent)


def indented_children(
    children: Sequence[RenderableAndLineAnnotations | RenderableTreePart],
    comma_separated: bool = False,
    force_trailing_comma: bool = False,
) -> RenderableTreePart:
  """Builds a IndentedChildren instance, supporting annotations and delimiters.

  This method stacks the children together, optionally inserting delimiters,
  and moving any comments to the end of their lines.

  Args:
    children: Children to render.
    comma_separated: Whether to automatically insert commas between children. If
      False, delimiters can be manually inserted into `children` first instead.
    force_trailing_comma: Whether to render a trailing comma in collapsed mode.

  Returns:
    A renderable part that renders the children on separate lines with an
    indent. When collapsed, it instead concatenates them.
  """
  return IndentedChildren.build(
      children=children,
      comma_separated=comma_separated,
      force_trailing_comma=force_trailing_comma,
  )


@dataclasses.dataclass(frozen=True)
class StyledBoxWithOutline(RenderableTreePart, abc.ABC):
  """An outlined box, which displays in "block" mode when rendered to HTML.

  Outlined boxes ensure that their child appears as a contiguous chunk, instead
  of having its first line indented, so that it can be fully encapsulated in
  a box.

  When rendered to HTML, this class may or may not insert extra newlines before
  and after the child, depending on whether this child was already alone on its
  line. When rendered to text, we always insert extra comments above and below
  the line.

  To allow the box to be collapsed separately, consider wrapping it in a
  foldable node.

  Attributes:
    child: The child to render.
    css_style: The style for the inner box.
  """

  child: RenderableTreePart
  css_style: str

  # Implementations of parent class abstract methods.
  def _compute_collapsed_width(self) -> int:
    return self.child.collapsed_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    # Conservatively assume we will insert newlines when expanded
    # (may or may not be true).
    count = 2 + self.child.newlines_in_expanded_parent
    return count

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return self.child.layout_marks_in_this_part

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return self.child.foldables_in_this_part()

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    base_rule = html_escaping.without_repeated_whitespace(f"""
        .outerbox_for_outline
        {{
            display: inline-block;
            padding: 0.25em;
            box-sizing: border-box;
        }}
        .box_with_outline
        {{
            display: inline-block;
            padding: 0.25em;
            box-sizing: border-box;
        }}
        .box_with_outline:not({context.collapsed_selector} *),
        .outerbox_for_outline:not({context.collapsed_selector} *)
        {{
            display: block;
            width: max-content;
        }}
        """)
    return {CSSStyleRule(base_rule)} | self.child.html_setup_parts(context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    style = html_escaping.escape_html_attribute(self.css_style)
    stream.write(
        '<span class="outerbox_for_outline"><span class="box_with_outline" '
        f'style="{style}">'
    )
    # Always treat the child as if it is at the beginning of the line, since it
    # will be once the box is expanded.
    self.child.render_to_html(
        stream,
        at_beginning_of_line=True,
        render_context=render_context,
    )
    stream.write("</span></span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # Insert separators before and after each child if in expanded mode.
    boxwidth = 40
    if expanded_parent:
      stream.write("\n" + " " * indent + "#╭" + "┄" * boxwidth + "╮\n")
      stream.write("\n" + " " * indent)

    self.child.render_to_text(
        stream,
        expanded_parent=expanded_parent,
        indent=indent,
        roundtrip_mode=roundtrip_mode,
        render_context=render_context,
    )

    if expanded_parent:
      stream.write("\n" + " " * indent + "#╰" + "┄" * boxwidth + "╯")
      stream.write("\n" + " " * indent)


def in_outlined_box(
    child: part_interface.RenderableTreePart,
    css_style: str = "outline: 1px dashed #aaaaaa;",
) -> RenderableTreePart:
  """Wraps a child into an outlined box.

  Outlined boxes ensure that their child appears as a contiguous chunk, instead
  of having its first line indented, so that it can be fully encapsulated in
  a box.

  When rendered to HTML, this class may or may not insert extra newlines before
  and after the child, depending on whether this child was already alone on its
  line. When rendered to text, we always insert extra comments above and below
  the line.

  To allow the box to be collapsed separately, consider wrapping it in a
  foldable node.

  Args:
    child: The child to render.
    css_style: The CSS style for the box element. By default, renders with a
      dashed grey outline.

  Returns:
    A renderable part that renders the child inside a standalone box.
  """
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  if not isinstance(css_style, str):
    raise ValueError(
        f"`css_style` must be a string, but got {type(css_style).__name__}"
    )
  return StyledBoxWithOutline(child, css_style)


@dataclasses.dataclass(frozen=True)
class WithHoverTooltip(DeferringToChild):
  """Customized renderer that shows extra info on hover.

  Attributes:
    child: Child to render.
    tooltip: Text to show when hovered over with the mouse.
  """

  child: RenderableTreePart
  tooltip: str

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .has_hover_tooltip {
        text-decoration: underline dotted;
        position: relative;
      }
      .has_hover_tooltip:hover::after  {
        display: block;
        position: absolute;
        top: calc(100% + 0.2ch);
        left: 0.2ch;
        content: attr(data-tooltip);
        outline: 1px dashed oklch(30% 0 0);
        background-color: oklch(90% 0 0);
        color: oklch(30% 0 0);
        z-index: 100;
        padding: 0.2ch;
        white-space: pre;
      }
    """))
    return {rule} | self.child.html_setup_parts(context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    tooltip_attr = html_escaping.escape_html_attribute(self.tooltip)
    stream.write(
        f'<span class="has_hover_tooltip" data-tooltip="{tooltip_attr}">'
    )
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


@dataclasses.dataclass(frozen=True)
class ScopedSelectableAnnotation(DeferringToChild):
  """Modifies its child so that selections outside it don't include it.

  It is sometimes useful to add additional annotations to an object that aren't
  pretty-printed parts of that object. This causes some problems for ordinary
  roundtrip mode, since we want it to be possible to exactly round-trip an
  object based on its printed representation.

  This object marks its child so that it doesn't get selected by the mouse
  when the selection starts outside the node. This means that if you copy the
  object normally, you don't copy the annotation, so that what you copied
  stays roundtrippable.

  In text mode, selections can't be manipulated. We fake the same thing by
  rendering it as a "comment" (by adding comment markers before every line)
  and hiding it if collapsed.

  Attributes:
    child: Child to render.
  """

  child: part_interface.RenderableTreePart

  def html_setup_parts(
      self, context: part_interface.HtmlContextForSetup
  ) -> set[part_interface.CSSStyleRule | part_interface.JavaScriptDefn]:
    # We want to ensure that the non-roundtrippable child does not get copied
    # if the user is trying to copy the roundtrippable version of the tree.
    # However, if the user wants to copy part of the non-roundtrippable child
    # itself, we don't want to get in the way.
    # We accomplish this by using the HTML/CSS "focus" mechanism, which the
    # browser automatically handles for anything with a "tabindex" attribute.
    # Specifically, when the user clicks on the rendered object, that object
    # gains focus, and when they click outside it (e.g. on a parent of this
    # object), the object looses focus. We can then disable text selection
    # when the object doesn't have focus, which means that if a text selection
    # starts outside the object, it won't include anything inside the object.
    rule = html_escaping.without_repeated_whitespace("""
        .scoped_unselectable:not(:focus)
        {
            user-select: none;
        }
    """)

    return {part_interface.CSSStyleRule(rule)} | self.child.html_setup_parts(
        context
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    # We need to set the "tabindex" attribute to make the text selection
    # adjustment work. -1 means that this should be selectable by the mouse
    # but should not be selectable by pressing "tab" on the keyboard or using
    # assistive tools.
    stream.write('<span tabindex="-1" class="scoped_unselectable">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    if roundtrip_mode:
      if not expanded_parent:
        # Don't render anything when collapsed and in roundtrip mode.
        return
      local_stream = io.StringIO()
      self.child.render_to_text(
          local_stream,
          expanded_parent=expanded_parent,
          indent=0,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )
      for line in local_stream.getvalue().splitlines():
        if line:
          stream.write("\n" + " " * indent + "# ")
          stream.write(line)
        else:
          stream.write("\n")
      stream.write("\n" + " " * indent)
    else:
      self.child.render_to_text(
          stream,
          expanded_parent=expanded_parent,
          indent=indent,
          roundtrip_mode=roundtrip_mode,
          render_context=render_context,
      )


def floating_annotation_with_separate_focus(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Wraps a child so that selections outside it don't include it.

  It is sometimes useful to add additional annotations to an object that aren't
  pretty-printed parts of that object. This causes some problems for ordinary
  roundtrip mode, since we want it to be possible to exactly round-trip an
  object based on its printed representation.

  This object marks its child so that it doesn't get selected by the mouse
  when the selection starts outside the node. This means that if you copy the
  object normally, you don't copy the annotation, so that what you copied
  stays roundtrippable.

  In text mode, selections can't be manipulated. We fake the same thing by
  rendering it as a "comment" (by adding comment markers before every line)
  and hiding it if collapsed.

  Args:
    child: Child to render.

  Returns:
    A wrapped version of ``child`` that does not participate in text selection
    in HTML mode, and adds comments in text mode.
  """
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return ScopedSelectableAnnotation(child)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Treescope: An interactive HTML pretty-printer and array visualizer.

You can configure treescope as the default IPython pretty-printer using ::

  treescope.basic_interactive_setup()

or, for more control: ::

  treescope.register_as_default()
  treescope.register_autovisualize_magic()
  treescope.active_autovisualizer.set_globally(
      treescope.ArrayAutovisualizer()
  )

You can also pretty-print individual values using `treescope.show` or
`treescope.display`.
"""
from __future__ import annotations

# pylint: disable=g-importing-member,g-multiple-import,unused-import

import typing

from . import _internal
from . import canonical_aliases
from . import context
from . import dataclass_util
from . import figures
from . import formatting_util
from . import handlers
from . import lowering
from . import ndarray_adapters
from . import renderers
from . import rendering_parts
from . import repr_lib
from . import type_registries

from ._internal.api.array_autovisualizer import (
    ArrayAutovisualizer,
)
from ._internal.api.arrayviz import (
    default_diverging_colormap,
    default_sequential_colormap,
    integer_digitbox,
    render_array,
    render_array_sharding,
)
from ._internal.api.autovisualize import (
    IPythonVisualization,
    VisualizationFromTreescopePart,
    ChildAutovisualizer,
    Autovisualizer,
    active_autovisualizer,
)
from ._internal.api.default_renderer import (
    active_renderer,
    active_expansion_strategy,
    render_to_html,
    render_to_text,
    using_expansion_strategy,
)
from ._internal.api.ipython_integration import (
    default_magic_autovisualizer,
    basic_interactive_setup,
    display,
    register_as_default,
    register_autovisualize_magic,
    register_context_manager_magic,
    show,
)

if typing.TYPE_CHECKING:
  Callable = typing.Callable

# Type annotations and docstrings for imported constants. Used by sphinx to
# generate documentation.

default_diverging_colormap: context.ContextualValue[
    list[tuple[int, int, int]]
] = default_diverging_colormap
"""Default diverging colormap.

Used by `render_array` when ``around_zero`` is True. Intended for user
customization in an interactive setting.
"""

default_sequential_colormap: context.ContextualValue[
    list[tuple[int, int, int]]
] = default_sequential_colormap
"""Default sequential colormap.

Used by `render_array` when ``around_zero`` is False. Intended for user
customization in an interactive setting.
"""

active_autovisualizer: context.ContextualValue[Autovisualizer | None] = (
    active_autovisualizer
)
"""The active autovisualizer to use when rendering a tree to HTML.

This can be overridden interactively to enable rich visualizations in
treescope. Users are free to set this to an arbitrary renderer of their
choice; a common choice is arrayviz's `ArrayAutovisualizer()`.
"""

active_renderer: context.ContextualValue[renderers.TreescopeRenderer] = (
    active_renderer
)
"""The default renderer to use when rendering a tree to HTML.

This determines the set of handlers and postprocessors to use when
rendering an object.

This can be overridden locally to reconfigure how nodes are rendered
with treescope. Users are free to set this to an arbitrary renderer of
their choice, and programs should not assume the renderer has a
particular form. Library functions can retrieve the current value of
this to render objects in a user-configurable way, and can optionally
make further adjustments using `TreescopeRenderer.extend_with`.
"""

active_expansion_strategy: context.ContextualValue[
    Callable[[rendering_parts.RenderableTreePart], None]
] = active_expansion_strategy
"""The default expansion strategy to use when rendering a tree to HTML.

Expansion strategies are used to figure out how deeply to unfold an object
by default. They should operate by setting the expand states of the
foldable nodes inside the object.
"""

default_magic_autovisualizer: context.ContextualValue[Autovisualizer] = (
    default_magic_autovisualizer
)
"""The default autovisualizer to use for the ``%%autovisualize`` magic.

This can be overridden interactively to customize the autovisualizer
used by ``%%autovisualize``.
"""

# Package version.
__version__ = '0.1.6'


# Set up canonical aliases for the treescope API itself.
def _setup_canonical_aliases_for_api():
  import types  # pylint: disable=import-outside-toplevel

  for key, value in globals().items():
    if isinstance(value, (type, types.FunctionType)):
      canonical_aliases.add_alias(
          value, canonical_aliases.ModuleAttributePath(__name__, (key,))
      )


_setup_canonical_aliases_for_api()
del _setup_canonical_aliases_for_api, typing



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of certain "well-known" objects, such as module functions.

Taking the `repr` of a function or callable usually produces something like

  <function vmap at 0x7f98bf556440>

and in some cases produces something like

  <jax.custom_derivatives.custom_jvp object at 0x7f98c0b5f130>

This can make it hard to determine what object this actually is from a user
perspective, and it would likely be more user-friendly to just output `jax.vmap`
or `jax.nn.relu` as the representation of this object.

Many functions and classes store a reference to the location where they were
originally defined (in their __module__ and __qualname__ attributes), which
can be used to find an alias for them. However, this may not be the "canonical"
alias, because some modules re-export private symbols under a public namespace
(in particular, JAX and Equinox both do this).

This module contains a registry of canonical paths for specific functions and
other objects, so that they can be rendered in a predictable way by treescope.
The intended purpose is primarily for interactive printing and debugging,
although it also helps for reifying objects into executable code through
round-trippable pretty printing, which can enable a simple form of
serialization.

This module also supports walking the public API of a package to automatically
set aliases; this is done by default for JAX and a few other libraries to ensure
the pretty-printed outputs avoid private module paths whenever possible. This
is intended as a heuristic to construct readable aliases for common objects on a
best-effort basis. It is not guaranteed that these inferred aliases will always
be stable across different versions of the external libraries.
"""

import collections
import contextlib
import dataclasses
import inspect
import sys
import types
from typing import Any, Callable, Literal, Mapping
import warnings

from treescope import context


@dataclasses.dataclass(frozen=True)
class ModuleAttributePath:
  """Expected path where we can find a particular object in a module.

  Attributes:
    module_name: Fully-qualified name of the module (the key of the module in
      sys.modules) in which this object can be found.
    attribute_path: Sequence of attributes identifying this object, separated by
      dots ("."). For instance, if this is ["foo", "bar"] then we expect to find
      the object at "{module_name}.foo.bar".
  """

  module_name: str
  attribute_path: tuple[str, ...]

  def __str__(self):
    module_name = self.module_name
    attribute_path_str = "".join(f".{attr}" for attr in self.attribute_path)
    if module_name == "__main__":
      assert attribute_path_str.startswith(".")
      return attribute_path_str[1:]

    return f"{module_name}{attribute_path_str}"

  def retrieve(self, forgiving: bool = False) -> Any:
    """Retrieves the object at this path.

    Args:
      forgiving: If True, return None on failure instead of raising an error.

    Returns:
      The retrieved object, or None if it wasn't found and `forgiving` was True.

    Raises:
      ValueError: If the object wasn't found and `forgiving` is False.
    """
    if self.module_name not in sys.modules:
      if forgiving:
        return None
      raise ValueError(
          f"Invalid alias {self} pointing to a non-imported module"
          f" {self.module_name}"
      )
    try:
      the_module = sys.modules[self.module_name]
    except KeyError:
      if forgiving:
        return None
      else:
        raise

    current_object = the_module
    for attr in self.attribute_path:
      if hasattr(current_object, attr):
        current_object = getattr(current_object, attr)
      else:
        if forgiving:
          return None
        raise ValueError(
            f"Can't retrieve {self}: {self.module_name} does"
            f" not expose an attribute {self.attribute_path!r}."
        )

    return current_object


@dataclasses.dataclass(frozen=True)
class LocalNamePath:
  """Expected path where we can find a particular object in a local scope.

  The "local scope" can be any dictionary of values with string keys, but it
  is usually either the locals() or globals() for a particular scope (or the
  union of these). A local name path is only valid relative to a particular
  scope that was used to create it.

  Attributes:
    local_name: Name of the variable in the local scope that we are retrieving.
    attribute_path: Sequence of attributes identifying this object, separated by
      dots ("."). For instance, if this is ["foo", "bar"] then we expect to find
      the object at "{local_name}.foo.bar".
  """

  local_name: str
  attribute_path: tuple[str, ...]

  def __str__(self):
    attribute_path_str = "".join(f".{attr}" for attr in self.attribute_path)
    return f"{self.local_name}{attribute_path_str}"

  def retrieve(
      self, local_scope: dict[str, Any], forgiving: bool = False
  ) -> Any:
    """Retrieves the object at this path.

    Args:
      local_scope: The scope in which we should retrieve this value.
      forgiving: If True, return None on failure instead of raising an error.

    Returns:
      The retrieved object, or None if it wasn't found and `forgiving` was True.

    Raises:
      KeyError, AttributeError: If the object wasn't found and `forgiving` is
        False.
    """
    try:
      current_object = local_scope[self.local_name]
    except KeyError:
      if forgiving:
        return None
      else:
        raise

    for attr in self.attribute_path:
      try:
        current_object = getattr(current_object, attr)
      except AttributeError:
        if forgiving:
          return None
        else:
          raise

    return current_object


@dataclasses.dataclass(frozen=True)
class CanonicalAliasEnvironment:
  """An environment that defines a set of canonical aliases.

  Attributes:
    aliases: A mapping from id(some_object) to the path where we expect to find
      that object.
  """

  aliases: dict[int, ModuleAttributePath]


_alias_environment: context.ContextualValue[CanonicalAliasEnvironment] = (
    context.ContextualValue(
        module=__name__,
        qualname="_alias_environment",
        initial_value=CanonicalAliasEnvironment({}),
    )
)
"""The current environment for module-level canonical aliases.

All alias mutation and lookups occur relative to the current
environment. Usually, this will not need to be modified, but it can
be useful to modify for tests or other local modifications.
"""


def add_alias(
    the_object: Any,
    path: ModuleAttributePath,
    on_conflict: Literal["ignore", "overwrite", "warn", "error"] = "warn",
):
  """Adds an alias to this object.

  Args:
    the_object: Object to add an alias to.
    path: The path where we expect to find this object.
    on_conflict: What to do if we try to add an alias for an object that already
      has one.

  Raises:
    ValueError: If overwrite is False and this object already has an alias,
      or if this object is not accessible at this path.
  """
  alias_env = _alias_environment.get()
  if id(the_object) in alias_env.aliases:
    if on_conflict == "ignore":
      return
    elif on_conflict == "overwrite":
      pass  # Continue adding an alias
    elif on_conflict == "warn":
      warnings.warn(
          f"Not defining alias {path} for {the_object!r}: it already has an"
          f" alias {alias_env.aliases[id(the_object)]}."
      )
    elif on_conflict == "error":
      raise ValueError(
          f"Can't define alias {path} for {the_object!r}: it already has an"
          f" alias {alias_env.aliases[id(the_object)]}."
      )

  retrieved = path.retrieve()
  if retrieved is not the_object:
    raise ValueError(
        f"Can't define alias {path} for {the_object!r}: {path} "
        f" is a different object {retrieved!r}."
    )
  # OK, it's probably safe to add this object as a well-known alias.
  alias_env.aliases[id(the_object)] = path


_current_scope_for_local_aliases: context.ContextualValue[
    dict[int, LocalNamePath] | None
] = context.ContextualValue(
    module=__name__,
    qualname="_current_scope_for_local_aliases",
    initial_value=None,
)
"""A mapping from IDs to local names.

Should only be used or modified by `local_alias_names`.
"""


def lookup_alias(
    the_object: Any,
    infer_from_attributes: bool = True,
    allow_outdated: bool = False,
    allow_relative: bool = False,
) -> ModuleAttributePath | LocalNamePath | None:
  """Retrieves an alias for this object, if possible.

  This function checks if an alias has been registered for this object, and also
  makes sure the object is still available at that alias. It optionally also
  tries to infer a fallback alias using __module__ and __qualname__ attributes.

  Args:
    the_object: Object to get a well-known alias for.
    infer_from_attributes: Whether to use __module__ and __qualname__ attributes
      to infer an alias if no explicit path is registered.
    allow_outdated: If True, return old aliases even if the object is no longer
      accessible at this path. (For instance, if the module was reloaded after
      this class/function was defined.)
    allow_relative: If True, return aliases that are local to the current
      `relative_alias_names` context if possible.

  Returns:
    A path at which we can find this object, or None if we do not have a path
    for it (or if the path is no longer correct).
  """
  if the_object is None:
    # None itself should never have an alias. Checking for it here lets us
    # easily catch the "broken alias" case in `retrieve` below.
    return None

  alias_env = _alias_environment.get()
  # Is this object in the local aliases? If so, return that.
  if allow_relative:
    local_aliases = _current_scope_for_local_aliases.get()
  else:
    local_aliases = None

  if local_aliases and id(the_object) in local_aliases:
    return local_aliases[id(the_object)]

  # Check for a global canonical alias.
  if id(the_object) in alias_env.aliases:
    alias = alias_env.aliases[id(the_object)]
  elif infer_from_attributes:
    # Try to unwrap it, in case it's a function-like object wrapping a function.
    unwrapped = inspect.unwrap(the_object)
    if (
        hasattr(unwrapped, "__module__")
        and hasattr(unwrapped, "__qualname__")
        and unwrapped.__qualname__ is not None
        and "<" not in unwrapped.__qualname__
    ):
      alias = ModuleAttributePath(
          unwrapped.__module__, tuple(unwrapped.__qualname__.split("."))
      )
    elif isinstance(the_object, types.ModuleType):
      alias = ModuleAttributePath(the_object.__name__, ())
    else:
      # Can't infer an alias.
      return None
  else:
    return None

  if not allow_outdated:
    if isinstance(the_object, types.MethodType):
      # Methods get different IDs on each access.
      if alias.retrieve(forgiving=True) != the_object:
        return None
    else:
      if alias.retrieve(forgiving=True) is not the_object:
        return None

  if local_aliases:
    # Check if any of the attributes along this path have a local alias.
    for split_point in reversed(range(len(alias.attribute_path))):
      parent_object = ModuleAttributePath(
          alias.module_name, alias.attribute_path[:split_point]
      ).retrieve(forgiving=True)
      if id(parent_object) in local_aliases:
        local_path_to_parent = local_aliases[id(parent_object)]
        return LocalNamePath(
            local_name=local_path_to_parent.local_name,
            attribute_path=(
                local_path_to_parent.attribute_path
                + alias.attribute_path[split_point:]
            ),
        )

    # Check if any parent module of the module in which this global alias is
    # defined has a local alias that we should use instead (e.g. using `np`
    # instead of `numpy`).
    module_parts = alias.module_name.split(".")
    submodule_path_reversed = []
    while module_parts:
      # Get an equivalent path from the parent module.
      parent_module_alias = ModuleAttributePath(
          ".".join(module_parts),
          tuple(reversed(submodule_path_reversed)) + alias.attribute_path,
      )
      # Is this a real parent module? (It probably should be unless someone
      # mucked with import paths.)
      if parent_module_alias.module_name in sys.modules:
        module_id = id(sys.modules[parent_module_alias.module_name])
        # Do we have a local alias for this parent module? And, if so, can we
        # access this particular object through the parent module in the
        # expected way?
        if (
            module_id in local_aliases
            and parent_module_alias.retrieve(forgiving=True) is the_object
        ):
          # First lookup the module in the local scope, then lookup the value
          # in the module.
          local_path_to_module = local_aliases[module_id]
          return LocalNamePath(
              local_name=local_path_to_module.local_name,
              attribute_path=(
                  local_path_to_module.attribute_path
                  + parent_module_alias.attribute_path
              ),
          )
      submodule_path_reversed.append(module_parts.pop())

  return alias


def maybe_local_module_name(module: types.ModuleType) -> str:
  """Returns a name for this module, possibly looking up local aliases."""
  alias = lookup_alias(module, allow_outdated=True, allow_relative=True)
  assert alias is not None
  return str(alias)


def default_well_known_filter(
    the_object: Any, path: ModuleAttributePath | LocalNamePath
) -> bool:
  """Checks if an object looks like something we want to define an alias for."""

  is_function_after_unwrap = isinstance(
      inspect.unwrap(the_object), types.FunctionType
  )

  # Only define aliases for objects that are either (a) mutable or (b)
  # classes/functions/modules.
  if (
      (hasattr(the_object, "__hash__") and the_object.__hash__ is not None)
      and not isinstance(the_object, (types.ModuleType, type))
      and not is_function_after_unwrap
  ):
    return False

  # Don't allow classes and functions to be assigned to any name other than the
  # name they were given at creation time.
  if isinstance(the_object, type) or (
      is_function_after_unwrap and hasattr(the_object, "__name__")
  ):
    expected_name = the_object.__name__
    if path.attribute_path and expected_name != path.attribute_path[-1]:
      return False

  # Assume any name that starts with an underscore is private.
  if any(attr.startswith("_") for attr in path.attribute_path) or (
      isinstance(path, LocalNamePath) and path.local_name.startswith("_")
  ):
    return False

  return True


def relative_alias_names(
    relative_scope: Mapping[str, Any] | Literal["magic"],
    predicate: Callable[[Any, LocalNamePath], bool] = default_well_known_filter,
) -> contextlib.AbstractContextManager[None]:
  """Context manager that makes `lookup_alias` return relative aliases.

  Args:
    relative_scope: A dictionary mapping in-scope names to their values, e.g.
      `globals()` or `{**globals(), **locals()}`. Objects that are reachable
      from this scope will have aliases that reference the keys in this dict.
      Alternatively, can be the string "magic", in which case we will walk the
      stack and use the `{**globals(), **locals()}` of the caller. ("magic" is
      only recommended for interactive usage.)
    predicate: A filter function to check if an object should be given an alias.

  Returns:
    A context manager in which `lookup_alias` will try to return relative
    aliases when `allow_relative=True`.
  """
  if relative_scope == "magic":
    # Infer the scope by walking the stack. 1 is the caller's frame.
    caller_frame_info = inspect.stack()[1]
    relative_scope = collections.ChainMap(
        caller_frame_info.frame.f_globals,
        caller_frame_info.frame.f_locals,
    )

  local_alias_map = {}

  assert isinstance(relative_scope, Mapping)
  for key, value in relative_scope.items():
    path = LocalNamePath(key, ())
    if predicate(value, path):
      local_alias_map[id(value)] = path

  return _current_scope_for_local_aliases.set_scoped(local_alias_map)


def populate_from_public_api(
    module: types.ModuleType,
    predicate: Callable[
        [Any, ModuleAttributePath], bool
    ] = default_well_known_filter,
):
  """Populates canonical aliases with all public symbols in a module.

  Attempts to walk this module and its submodules to extract well-known symbols.
  Symbols that already have an alias defined will be ignored.

  If the module defines __all__, we assume the symbols in __all__ are the
  well-known symbols in this module. (See
  https://docs.python.org/3/reference/simple_stmts.html#the-import-statement)

  If the module does not define __all__, we look for all names that do not
  start with "_".

  We then additionally filter down this set to the set of objects for which
  the `predicate` argument returns True.

  This function should only be called on modules with a well-defined public API,
  that exports only functions defined in the module itself. If a module
  re-exports symbols from another external module (e.g. importing `partial`
  from `functools`), we might otherwise end up using the unrelated module as
  the "canonical" source of that object. (The `prefix_filter` function below
  tries to prevent this if possible when used as a predicate.)

  Args:
    module: The module we will collect symbols from.
    predicate: A filter function to check if an object should be given an alias.
  """
  if hasattr(module, "__all__"):
    public_names = module.__all__
  else:
    public_names = [
        key for key in module.__dict__.keys() if not key.startswith("_")
    ]

  for name in public_names:
    try:
      value = getattr(module, name)
    except AttributeError:
      # Possibly a misspecified __all__?
      continue
    path = ModuleAttributePath(module.__name__, (name,))
    if isinstance(value, types.ModuleType):
      if (
          value.__name__.startswith(module.__name__)
          and value.__name__ != module.__name__
      ):
        # Process submodules of this module also.
        populate_from_public_api(value, predicate)
      # Don't process external modules that are being re-exported.
    elif predicate(value, path):
      add_alias(value, path, on_conflict="ignore")


def prefix_filter(include: str, excludes: tuple[str, ...] = ()):
  """Builds a filter that only defines aliases within a given prefix."""

  def is_under_prefix(path: str, prefix: str):
    return path == prefix or path.startswith(prefix + ".")

  def predicate(the_object: Any, path: ModuleAttributePath) -> bool:
    if not default_well_known_filter(the_object, path):
      return False
    if not is_under_prefix(str(path), include):
      return False
    if any(is_under_prefix(str(path), exclude) for exclude in excludes):
      return False
    if (
        hasattr(the_object, "__module__")
        and the_object.__module__
        and not is_under_prefix(the_object.__module__, include)
    ):
      return False
    return True

  return predicate



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Algorithms for expanding and collapsing foldable nodes."""

from __future__ import annotations

import collections
from typing import Any, Collection

from treescope._internal.parts import part_interface


def expand_to_layout_marks(
    tree: part_interface.RenderableTreePart,
    marks: Collection[Any],
    collapse_weak_others: bool,
) -> None:
  """Expands a tree so that any part with a marker in the given set is visible.

  This function ignores the existing expand state, and instead just ensures that
  every foldable node that contains a layout marker in `marks` is expanded.
  Optionally it will also ensure that all sibling foldable nodes are collapsed.

  Expand states for those nodes are set to EXPANDED or COLLAPSED, and expand
  states for other nodes are not modified. This means you can call
  `expand_for_balanced_layout` after this function to reformat the subtrees
  of `tree` that haven't yet been assigned strong expand/collapse states. This
  is useful for producing balanced layouts once the user expands a collapsed
  node.

  Args:
    tree: The tree to update.
    marks: The marks that should be made visible.
    collapse_weak_others: Whether to collapse foldables that do NOT have the
      given marks and have a weak expansion state.
  """
  for foldable in tree.foldables_in_this_part():
    found = _process_foldable_by_marks(foldable, marks, collapse_weak_others)
    if collapse_weak_others and not found:
      foldable.set_expand_state(part_interface.ExpandState.COLLAPSED)


def _process_foldable_by_marks(
    foldable: part_interface.FoldableTreeNode,
    marks: Collection[Any],
    collapse_weak_others: bool,
) -> bool:
  """Expands a foldable based on marks, and may also collapse its children.

  Args:
    foldable: The foldable to process.
    marks: The marks that should be made visible.
    collapse_weak_others: Whether to collapse foldables that do NOT have the
      given marks and have a weak expansion state.

  Returns:
    True if this node should be made visible (e.g. all parents expanded) based
    on the marks.
  """
  # Check children.
  might_need_collapsing = []
  for child_foldable in foldable.as_expanded_part().foldables_in_this_part():
    if not _process_foldable_by_marks(
        child_foldable, marks, collapse_weak_others
    ):
      might_need_collapsing.append(child_foldable)

  if any(mark in foldable.layout_marks_in_this_part for mark in marks):
    # We need to expand this node.
    foldable.set_expand_state(part_interface.ExpandState.EXPANDED)
    if collapse_weak_others:
      # We should also collapse any child that wasn't marked as needing to be
      # expanded.
      for child_foldable in might_need_collapsing:
        if child_foldable.get_expand_state().is_weak():
          child_foldable.set_expand_state(part_interface.ExpandState.COLLAPSED)
    # Inform caller that we found something.
    return True
  else:
    # This node doesn't need to be expanded. Don't immediately mark it as
    # collapsed, because we may want to mark a parent as collapsed instead.
    # The parent will collapse this node if necessary.
    return False


def expand_for_balanced_layout(
    tree: part_interface.RenderableTreePart,
    max_height: int | None = 20,
    target_width: int = 60,
    relax_height_constraint_for_only_child: bool = True,
    recursive_expand_height_for_collapsed_nodes: int = 7,
) -> None:
  """Updates a tree's (weak) expand states to achieve a "balanced" layout.

  A balanced layout tries to evenly allocate space along vertical and horizontal
  axes. Similar to many pretty-printers, nodes are usually expanded to avoid
  long lines and make them fit comfortably within `target_width` characters.
  However, there is also a `max_height` parameter that takes precedence, which
  can be used to prevent outputs from being too large for very large objects,
  and makes it easier for users to interactively expand only the subtrees that
  they want to inspect.

  Unlike other pretty-printers, for simplicity we don't track the current
  indent level or remaining line width, but instead make local decisions for
  each node. This means the overall rendering may end up longer than
  `target_width` because the leaves (each of width `target_width`) may be
  indented. (This simplifies the system because we do not have to track the
  exact indent levels and starting/ending positions of each node, and don't
  have to propagate that information up through the rendered parts.)

  This function only modifies WEAKLY_EXPANDED or WEAKLY_COLLAPSED nodes.

  Args:
    tree: The tree to update.
    max_height: Maximum number of lines for the rendering, which we should try
      not to exceed when expanding nodes. Can sometimes be exceeded for "only
      child" subtrees if `relax_height_constraint_for_only_child` is set. If
      None, then an arbitrary height is allowed, and only `target_width` will be
      used to determine expansion size.
    target_width: Target number of characters for the rendering. If a node is
      shorter than this width when collapsed, we will leave it on a single line
      instead of expanding it.
    relax_height_constraint_for_only_child: If True, we allow the height
      constraint to be violated once if there is only a single node that could
      possibly be expanded. In particular, we repeatedly expand nodes until we
      reach a node shorter than `target_width` or a node with two or more
      children. If we match or exceed the height constraint while doing this, we
      stop after expanding that node (but leave it expanded). If we haven't
      reached the constraint, we continue expanding under the stricter rules.
    recursive_expand_height_for_collapsed_nodes: Whenever we mark a node as
      collapsed, we first recursively expand its children as if it was expanded
      with this as the max height, then collapse only the outermost node. This
      means it will render as a single line, but if expanded by the user, it
      will expand not just the clicked node but also some of its children. This
      is intended to support faster interactive exploration of large objects.
  """
  # Tracks the number of lines we have so far committed to outputting via our
  # expansion decisions.
  committed_height = 1 + tree.newlines_in_expanded_parent
  # Tracks the collection of foldable nodes we can consider expanding.
  candidate_foldables = collections.deque(tree.foldables_in_this_part())

  # Step 1: Handle only children.
  if relax_height_constraint_for_only_child:
    while (
        # We haven't yet reached the height constraint.
        (max_height is None or committed_height < max_height)
        # There's only one node that could be expanded.
        and len(candidate_foldables) == 1
        # This node doesn't fit in the target width.
        and candidate_foldables[0].collapsed_width > target_width
        # This node isn't already forced to be collapsed.
        and candidate_foldables[0].get_expand_state()
        != part_interface.ExpandState.COLLAPSED
    ):
      # We should expand this node.
      foldable = candidate_foldables.popleft()
      expanded_foldable_contents = foldable.as_expanded_part()
      foldable.set_expand_state(part_interface.ExpandState.EXPANDED)
      committed_height += expanded_foldable_contents.newlines_in_expanded_parent
      candidate_foldables.extend(
          expanded_foldable_contents.foldables_in_this_part()
      )

  # Step 2: Expand nodes in breadth-first order while we have space.
  while candidate_foldables:
    foldable = candidate_foldables.popleft()
    expanded_foldable_contents = foldable.as_expanded_part()

    if foldable.get_expand_state().is_weak():
      # Infer a state for this node.
      height_if_expanded = (
          committed_height
          + expanded_foldable_contents.newlines_in_expanded_parent
      )
      if max_height is not None and height_if_expanded > max_height:
        # This node takes up too much space if expanded. Collapse it.
        foldable.set_expand_state(part_interface.ExpandState.COLLAPSED)
      elif foldable.collapsed_width > target_width:
        # This node is too long to render on one line. Expand it.
        foldable.set_expand_state(part_interface.ExpandState.EXPANDED)
      else:
        # This node could render either collapsed or expanded. Commit to its
        # current weak preference.
        if (
            foldable.get_expand_state()
            == part_interface.ExpandState.WEAKLY_EXPANDED
        ):
          foldable.set_expand_state(part_interface.ExpandState.EXPANDED)
        else:
          foldable.set_expand_state(part_interface.ExpandState.COLLAPSED)

    if foldable.get_expand_state() == part_interface.ExpandState.EXPANDED:
      # Record it as expanded and add children to our queue.
      committed_height += expanded_foldable_contents.newlines_in_expanded_parent
      candidate_foldables.extend(
          expanded_foldable_contents.foldables_in_this_part()
      )
    else:
      # Recursively process the children with this function, and don't add them
      # to the current queue (since their parent isn't expanded)
      expand_for_balanced_layout(
          expanded_foldable_contents,
          max_height=recursive_expand_height_for_collapsed_nodes,
          target_width=target_width,
          relax_height_constraint_for_only_child=(
              relax_height_constraint_for_only_child
          ),
          recursive_expand_height_for_collapsed_nodes=(
              recursive_expand_height_for_collapsed_nodes
          ),
      )


def expand_to_depth(
    tree: part_interface.RenderableTreePart, depth: int
) -> None:
  """Expands a tree up to a given depth.

  This function ignores the existing expand state, and instead rewrites it to
  expand up to the given depth.

  Args:
    tree: Tree to expand.
    depth: Depth to expand to. At depth 0, all foldables will be collapsed.
  """
  for foldable in tree.foldables_in_this_part():
    if depth > 0:
      foldable.set_expand_state(part_interface.ExpandState.EXPANDED)
      expand_to_depth(foldable.as_expanded_part(), depth - 1)
    else:
      foldable.set_expand_state(part_interface.ExpandState.COLLAPSED)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for external (non-treescope) library types.

Submodules of this module are not always imported by default. Instead, they are
dynamically imported and configured when Treescope detects that the
corresponding external library has been imported. (See
`treescope.type_registries`.)
"""



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""User-configurable automatic visualization of subtrees.

This module provides a quick but flexible interface for customizing the
output of Treescope by replacing subtrees by rendered figures. The intended
use case is to make it easy to generate visualizations for arrays, parameters,
or other information, even if they are deeply nested inside larger data
structures.

This module also defines a particular automatic visualizer that uses Arrayviz
to visualize all arrays in a tree, which can be useful for exploratory analysis
of models and data.
"""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Protocol

from treescope import context
from treescope import rendering_parts
from treescope._internal import object_inspection


@dataclasses.dataclass
class IPythonVisualization:
  """Used by autovisualizers to replace a subtree with a display object.

  Attributes:
    display_object: An object to render, e.g. a figure.
    replace: Whether to replace the subtree completely with this display
      object. If False, the display object will be rendered below the object.
  """

  display_object: object_inspection.HasReprHtml
  replace: bool = False


@dataclasses.dataclass
class VisualizationFromTreescopePart:
  """Used by advanced autovisualizers to directly insert Treescope content.

  Attributes:
    rendering: A custom treescope rendering which will completely replace the
      subtree.
  """

  rendering: rendering_parts.RenderableAndLineAnnotations


@dataclasses.dataclass
class ChildAutovisualizer:
  """Used by autovisualizers to switch to a different autovisualizer.

  This can be used to enable conditional logic, e.g. "when rendering an object
  of type Foo, visualize objects of type Bar". Note that the returned child
  autovisualizer is also allowed to close over things from the current call,
  e.g. you can return a lambda function that checks if it is rendering a
  particular child of the original node.

  Attributes:
    autovisualizer: The new autovisualizer to use while rendering this value's
      children.
  """

  autovisualizer: Autovisualizer


class Autovisualizer(Protocol):
  """Protocol for autovisualizers.

  An autovisualizer is a callable object that can be used to customize
  Treescope for rendering purposes. The active autovisualizer is called once
  for each subtree of a rendered PyTree, and can choose to:

  * render it normally (but possibly customize its children),

  * replace the rendering of that PyTree with a custom rich display object,

  * or replace the current autovisualizer with a new one that will be active
    while rendering the children.
  """

  @abc.abstractmethod
  def __call__(
      self, value: Any, path: str | None
  ) -> (
      IPythonVisualization
      | VisualizationFromTreescopePart
      | ChildAutovisualizer
      | None
  ):
    """Runs the autovisualizer on a node at a given path.

    Args:
      value: A value being rendered in treescope.
      path: Optionally, a path to this node, represented as a string that can be
        used to reach this node from the root (e.g. ".foo.bar['baz']").

    Returns:
      A visualization for this subtree, a child autovisualizer to use while
      rendering the node's children, or None to continue rendering the node
      normally using the current autovisualizer.
    """
    raise NotImplementedError()


active_autovisualizer: context.ContextualValue[Autovisualizer | None] = (
    context.ContextualValue(
        module=__name__,
        qualname="active_autovisualizer",
        initial_value=None,
    )
)
